#  Copyright (c) 2023 zfit
from __future__ import annotations

import pydantic
import yaml

from zfit.util.warnings import warn_experimental_feature


class ZfitSerializable:
    hs3_type: str = None

    @classmethod
    def get_repr(cls) -> pydantic.BaseModel:
        from zfit.serialization import Serializer

        return Serializer.type_repr[cls.hs3_type]

    def to_orm(self):
        raise NotImplementedError


class NumpyArrayNotSerializableError(TypeError):
    pass


class SerializableMixin(ZfitSerializable):
    hs3 = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.hs3 = type(self).hs3(self)

    def __init_subclass__(cls, **kwargs):
        cls.hs3 = create_HS3(cls)

    @warn_experimental_feature
    def to_yaml(self):
        """Convert the object to a yaml string.

        Returns:
            str: The yaml string.
        """
        json_obj = self.to_json()
        yaml_obj = yaml.safe_dump(json_obj)
        return yaml_obj

    @warn_experimental_feature
    def to_asdf(self):
        """Convert the object to an asdf file."""
        try:
            import asdf
        except ImportError:
            raise ImportError(
                "The asdf module is not installed. Please install zfit with the extra `hs3` (i.e. `pip install zfit[sh3]` or asdf directry to use this feature."
            )

        asdf_obj = asdf.AsdfFile(self.to_dict())
        return asdf_obj

    @classmethod
    @warn_experimental_feature
    def from_asdf(cls, asdf_obj):
        """Load an object from an asdf file."""

        asdf_tree = asdf_obj.tree
        # cleanup the asdf chunk
        asdf_tree.pop("asdf_library", None)
        asdf_tree.pop("history", None)
        orm = cls.from_dict(asdf_tree)
        return orm

    @warn_experimental_feature
    def to_json(self):
        """Convert the object to a json string.

        Returns:
            str: The json string.
        """
        from zfit.serialization import Serializer

        Serializer.initialize()
        repr = self.get_repr()
        orm = repr.from_orm(self)
        try:
            json_obj = orm.json(exclude_none=True, by_alias=True)
        except TypeError as error:
            if "Object of type 'ndarray' is not JSON serializable" in str(error):
                raise NumpyArrayNotSerializableError(
                    "The object you are trying to serialize contains numpy arrays. "
                    "This is not supported by json. Please use `to_asdf` (or `to_dict)` instead."
                )
            else:
                raise
        return json_obj

    @classmethod
    @warn_experimental_feature
    def from_json(cls, json: str) -> object:
        """Load an object from a json string.

        Args:
            json: Serialized object in a JSON string.

        Returns:
            The deserialized object.
        """
        from zfit.serialization import Serializer

        Serializer.initialize()
        orm = cls.get_repr().parse_raw(json)
        return orm.to_orm()

    @warn_experimental_feature
    def to_dict(self):
        """Convert the object to a nested dictionary structure.

        Returns:
               dict: The dictionary structure.
        """
        from zfit.serialization import Serializer

        Serializer.initialize()
        repr = self.get_repr()
        orm = repr.from_orm(self)
        return orm.dict(exclude_none=True, by_alias=True)

    @classmethod
    @warn_experimental_feature
    def from_dict(cls, dict_):
        """Creates an object from a dictionary structure as generated by `to_dict`.

        Args:
            dict_: Dictionary structure.

        Returns:
            The deserialized object.
        """
        from zfit.serialization import Serializer

        Serializer.initialize()
        orm = cls.get_repr().parse_obj(dict_)
        return orm.to_orm()

    @classmethod
    def get_repr(cls):
        """Abstract representation of the object for serialization.

        This objects knows how to serialize and deserialize the object and is used by the
        `to_json`, `from_json`, `to_dict` and `from_dict` methods.

        Returns:
            pydantic.BaseModel: The representation of the object.
        """
        try:
            from ..serialization import Serializer
        except ImportError:
            return None
        return Serializer.constructor_repr.get(cls)


class HS3:
    implementation = None

    def __init__(self, obj):
        super().__init__()
        self.obj = obj
        self.original_init = {}

    def to_json(self):
        orm = self.repr.from_orm(self)
        try:
            json_obj = orm.json(exclude_none=True, by_alias=True)
        except TypeError as error:
            if "TypeError: Object of type 'ndarray' is not JSON serializable" in str(
                error
            ):
                raise TypeError(
                    "The object you are trying to serialize contains numpy arrays. "
                    "This is not supported by json. Please use `to_asdf` (or `to_dict)` instead."
                )
            else:
                raise
        return json_obj

    def to_dict(self):
        orm = self.repr.from_orm(self)
        return orm.dict(exclude_none=True, by_alias=True)

    @classmethod
    def from_json(cls, json):
        orm = cls.implementation.get_repr().parse_raw(json)
        return orm.to_orm()

    @property
    def repr(self):
        return self.obj.get_repr()


def create_HS3(cls):
    class HS3Specialized(HS3):
        implementation = cls

    return HS3Specialized



"""This module contains functions for the numeric as well as the analytic (partial) integration."""

#  Copyright (c) 2024 zfit

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import zfit

from collections.abc import Callable

import collections
from contextlib import suppress
from typing import Union

import numpy as np
import tensorflow as tf
import tensorflow_probability as tfp

import zfit.z.numpy as znp
from zfit import z
from .dimension import BaseDimensional
from .interfaces import ZfitData, ZfitModel, ZfitSpace
from ..util.container import convert_to_container
from ..util.temporary import TemporarilySet

from ..settings import ztypes
from ..util import ztyping
from ..util.exception import AnalyticIntegralNotImplemented, WorkInProgressError
from .space import Space, convert_to_space, supports


def auto_integrate(
    func,
    limits,
    n_axes=None,
    x=None,
    method="AUTO",
    dtype=ztypes.float,
    mc_sampler=tfp.mcmc.sample_halton_sequence,
    max_draws=None,
    tol=None,
    vectorizable=None,
    mc_options=None,
    simpsons_options=None,
):
    if vectorizable is None:
        vectorizable = False
    limits = convert_to_space(limits)

    if n_axes is None:
        n_axes = limits.n_obs
    if method == "AUTO":  # TODO unfinished, other methods?
        if n_axes == 1 and x is None:
            method = "simpson"
        else:
            method = "mc"
    # TODO method
    if method.lower() == "mc":
        mc_options = mc_options or {}
        draws_per_dim = mc_options["draws_per_dim"]
        max_draws = mc_options.get("max_draws")
        integral = mc_integrate(
            x=x,
            func=func,
            limits=limits,
            n_axes=n_axes,
            method=method,
            dtype=dtype,
            mc_sampler=mc_sampler,
            draws_per_dim=draws_per_dim,
            max_draws=max_draws,
            tol=tol,
            importance_sampling=None,
            vectorizable=vectorizable,
        )
    elif method.lower() == "simpson":
        num_points = simpsons_options["draws_simpson"]
        integral = simpson_integrate(func=func, limits=limits, num_points=num_points)
    else:
        raise ValueError(f"Method {method} not a legal choice for integration method.")
    return integral


# TODO implement numerical integration method
def numeric_integrate():
    """Integrate ``func`` using numerical methods."""
    return None


# COPIED FROM tf_quant_finance/math/integrate.py which is deprecated
def simpson(func, lower, upper, num_points=1001, dtype=None, name=None):
    """Evaluates definite integral using composite Simpson's 1/3 rule.

    Integrates `func` using composite Simpson's 1/3 rule [1].

    Evaluates function at points of evenly spaced grid of `num_points` points,
    then uses obtained values to interpolate `func` with quadratic polynomials
    and integrates these polynomials.

    #### References
    [1] Weisstein, Eric W. "Simpson's Rule." From MathWorld - A Wolfram Web
        Resource. http://mathworld.wolfram.com/SimpsonsRule.html

    #### Example
    ```python
      f = lambda x: x*x
      a = tf.constant(0.0)
      b = tf.constant(3.0)
      integrate(f, a, b, num_points=1001) # 9.0
    ```

    Args:
      func: Python callable representing a function to be integrated. It must be a
        callable of a single `Tensor` parameter and return a `Tensor` of the same
        shape and dtype as its input. It will be called with a `Tesnor` of shape
        `lower.shape + [n]` (where n is integer number of points) and of the same
        `dtype` as `lower`.
      lower: `Tensor` or Python float representing the lower limits of
        integration. `func` will be integrated between each pair of points defined
        by `lower` and `upper`.
      upper: `Tensor` of the same shape and dtype as `lower` or Python float
        representing the upper limits of intergation.
      num_points: Scalar int32 `Tensor`. Number of points at which function `func`
        will be evaluated. Must be odd and at least 3.
        Default value: 1001.
      dtype: Optional `tf.Dtype`. If supplied, the dtype for the `lower` and
        `upper`. Result will have the same dtype.
        Default value: None which maps to dtype of `lower`.
      name: Python str. The name to give to the ops created by this function.
        Default value: None which maps to 'integrate_simpson_composite'.

    Returns:
      `Tensor` of shape `func_batch_shape + limits_batch_shape`, containing
        value of the definite integral.
    """
    with tf.compat.v1.name_scope(
        name, default_name="integrate_simpson_composite", values=[lower, upper]
    ):
        lower = tf.convert_to_tensor(lower, dtype=dtype, name="lower")
        dtype = lower.dtype
        upper = tf.convert_to_tensor(upper, dtype=dtype, name="upper")
        num_points = tf.convert_to_tensor(num_points, dtype=tf.int32, name="num_points")

        assertions = [
            tf.debugging.assert_greater_equal(num_points, 3),
            tf.debugging.assert_equal(num_points % 2, 1),
        ]

        with tf.compat.v1.control_dependencies(assertions):
            dx = (upper - lower) / (tf.cast(num_points, dtype=dtype) - 1)
            dx_expand = tf.expand_dims(dx, -1)
            lower_exp = tf.expand_dims(lower, -1)
            grid = lower_exp + dx_expand * tf.cast(tf.range(num_points), dtype=dtype)
            weights_first = tf.constant([1.0], dtype=dtype)
            weights_mid = tf.tile(
                tf.constant([4.0, 2.0], dtype=dtype), [(num_points - 3) // 2]
            )
            weights_last = tf.constant([4.0, 1.0], dtype=dtype)
            weights = tf.concat([weights_first, weights_mid, weights_last], axis=0)

        return tf.reduce_sum(func(grid) * weights, axis=-1) * dx / 3.0


def simpson_integrate(func, limits, num_points):  # currently not vectorized
    integrals = []
    num_points = tf.cast(num_points, znp.int32)
    num_points += num_points % 2 + 1  # sanitize number of points
    for space in limits:
        lower, upper = space.rect_limits
        if lower.shape[0] > 1:
            raise ValueError(
                "Vectorized spaces in integration currently not supported."
            )
        lower = znp.array(lower)[0, 0]
        upper = znp.array(upper)[0, 0]
        tf.debugging.assert_all_finite(
            (lower, upper),
            "MC integration does (currently) not support unbound limits (np.infty) as given here:"
            "\nlower: {}, upper: {}".format(lower, upper),
        )
        integrals.append(
            simpson(
                func=func,
                lower=lower,
                upper=upper,
                num_points=num_points,
                dtype=znp.float64,
            )
        )
    return znp.sum(integrals, axis=0)


# @z.function
def mc_integrate(
    func: Callable,
    limits: ztyping.LimitsType,
    axes: ztyping.AxesTypeInput | None = None,
    x: ztyping.XType | None = None,
    n_axes: int | None = None,
    draws_per_dim: int = 40000,
    max_draws=800_000,
    tol: float = 1e-6,
    method: str = None,
    dtype: type = ztypes.float,
    mc_sampler: Callable = tfp.mcmc.sample_halton_sequence,
    importance_sampling: Callable | None = None,
    vectorizable=None,
) -> tf.Tensor:
    """Monte Carlo integration of ``func`` over ``limits``.

    Args:
        vectorizable ():
        func: The function to be integrated over
        limits: The limits of the integral
        axes: The row to integrate over. None means integration over all value
        x: If a partial integration is performed, this are the value where x will be evaluated.
        n_axes: the number of total dimensions (old?)
        draws_per_dim: How many random points to draw per dimensions
        method: Which integration method to use
        dtype: |dtype_arg_descr|
        mc_sampler: A function that takes one argument (``n_draws`` or similar) and returns
            random value between 0 and 1.
        importance_sampling:

    Returns:
        The integral
    """
    import zfit

    if vectorizable is None:
        vectorizable = False
    tol = znp.array(tol, dtype=znp.float64)
    if axes is not None and n_axes is not None:
        raise ValueError("Either specify axes or n_axes")

    axes = limits.axes
    partial = (axes is not None) and (x is not None)  # axes, value can be tensors

    if axes is not None and n_axes is None:
        n_axes = len(axes)
    if n_axes is not None and axes is None:
        axes = tuple(range(n_axes))

    integrals = []
    for space in limits:
        lower, upper = space._rect_limits_tf
        tf.debugging.assert_all_finite(
            (lower, upper),
            "MC integration does (currently) not support unbound limits (np.infty) as given here:"
            "\nlower: {}, upper: {}".format(lower, upper),
        )

        n_samples = draws_per_dim * n_axes

        chunked_normalization = zfit.run.chunksize < n_samples
        # chunked_normalization = True
        if chunked_normalization and partial:
            print(
                "NOT SUPPORTED! partial and chunked not working, auto switch back to not-chunked."
            )
        if chunked_normalization and not partial:
            n_chunks = int(np.ceil(n_samples / zfit.run.chunksize))
            chunksize = int(np.ceil(n_samples / n_chunks))
            # print("starting normalization with {} chunks and a chunksize of {}".format(n_chunks, chunksize))
            avg = normalization_chunked(
                func=func,
                n_axes=n_axes,
                dtype=dtype,
                x=x,
                num_batches=n_chunks,
                batch_size=chunksize,
                space=space,
            )

        else:
            # TODO: deal with n_obs properly?
            @z.function(wraps="tensor", keepalive=True)
            def cond(avg, error, std, ntot, i):
                # return i < 3
                return znp.logical_and(error > tol, ntot < max_draws)

            def body_integrate(avg, error, std, ntot, i):
                ntot_old = ntot
                ntot += n_samples
                if partial:
                    samples_normed = tfp.mcmc.sample_halton_sequence(
                        dim=n_axes,
                        # sequence_indices=tf.range(ntot_old, ntot),
                        num_results=n_samples / 10,
                        # reduce, it explodes otherwise easily
                        # as we don't do adaptive now
                        # to decrease integration size
                        dtype=dtype,
                        randomized=False,
                    )
                else:
                    samples_normed = tfp.mcmc.sample_halton_sequence(
                        dim=n_axes,
                        sequence_indices=tf.range(ntot_old, ntot),
                        # num_results=n_samples,  # to decrease integration size
                        dtype=dtype,
                        randomized=False,
                    )
                samples = (
                    samples_normed * (upper - lower) + lower
                )  # samples is [0, 1], stretch it
                if partial:  # TODO(Mayou36): shape of partial integral?
                    data_obs = x.obs
                    new_obs = []
                    xval = x.value()
                    value_list = []
                    index_samples = 0
                    index_values = 0
                    if len(xval.shape) == 1:
                        xval = znp.expand_dims(xval, axis=1)
                    for i in range(n_axes + xval.shape[-1]):
                        if i in axes:
                            new_obs.append(space.obs[index_samples])
                            value_list.append(samples[:, index_samples])
                            index_samples += 1
                        else:
                            new_obs.append(data_obs[index_values])
                            value_list.append(
                                znp.expand_dims(xval[:, index_values], axis=1)
                            )
                            index_values += 1
                    value_list = [tf.cast(val, dtype=dtype) for val in value_list]
                    xval = PartialIntegralSampleData(
                        sample=value_list, space=Space(obs=new_obs)
                    )
                else:
                    xval = samples
                # convert rnd samples with value to feedable vector
                reduce_axis = 1 if partial else None
                y = func(xval)
                # avg = znp.mean(y)
                ifloat = tf.cast(i, dtype=tf.float64)
                if partial:
                    avg = znp.mean(y, axis=reduce_axis)
                else:
                    avg = avg / (ifloat + 1.0) * ifloat + znp.mean(
                        y, axis=reduce_axis
                    ) / (ifloat + 1.0)
                std = std / (ifloat + 1.0) * ifloat + znp.std(y) / (ifloat + 1.0)
                ntot_float = znp.asarray(ntot, dtype=znp.float64)

                # estimating the error of QMC is non-trivial
                # (https://www.degruyter.com/document/doi/10.1515/mcma-2020-2067/html or
                # https://stats.stackexchange.com/questions/533725/how-to-calculate-quasi-monte-carlo-integration-error-when-sampling-with-sobols)
                # However, we use here just something that is in the right direction for the moment being. Therefore,
                # we use the MC error as an upper bound as QMC is better/equal to MC (for our cases).
                error_sobol = std * znp.log(ntot_float) ** n_axes / ntot_float
                error_random = std / znp.sqrt(ntot_float)
                error = (
                    znp.minimum(error_sobol, error_random) * 0.1
                )  # heuristic factor from using QMC
                return avg, error, std, ntot, i + 1

            avg, error, std, ntot, i = [
                znp.array(0.0, dtype=znp.float64),
                znp.array(9999.0, dtype=znp.float64),  # init value large, irrelevant
                znp.array(0.0, dtype=znp.float64),
                0,
                0,
            ]
            if partial or vectorizable:
                avg, error, std, ntot, i = body_integrate(avg, error, std, ntot, i)
            else:
                avg, error, std, ntot, i = tf.while_loop(
                    cond=cond, body=body_integrate, loop_vars=[avg, error, std, ntot, i]
                )
                from zfit import settings

                if settings.get_verbosity() > 9:
                    tf.print("i:", i, "   ntot:", ntot)

            # avg = tfp.monte_carlo.expectation(f=func, samples=x, axis=reduce_axis)
            # TODO: importance sampling?
            # avg = tfb.monte_carlo.expectation_importance_sampler(f=func, samples=value,axis=reduce_axis)

            def print_none_return():
                from zfit import settings

                if settings.get_verbosity() >= 0:
                    tf.print(
                        "Estimated integral error (",
                        error,
                        ") larger than tolerance (",
                        tol,
                        "), which is maybe not enough (but maybe it's also fine)."
                        " You can (best solution) implement an anatytical integral (see examples in repo) or"
                        " manually set a higher number on the PDF with 'update_integration_options'"
                        " and increase the 'max_draws' (or adjust 'tol'). "
                        "If partial integration is chosen, this can lead to large memory consumption."
                        "This is a new warning checking the integral accuracy. It may warns too often as it is"
                        " Work In Progress. If you have any observation on it, please tell us about it:"
                        " https://github.com/zfit/zfit/issues/new/choose"
                        "To suppress this warning, use zfit.settings.set_verbosity(-1).",
                    )
                return

            if not vectorizable:
                tf.cond(error > tol, print_none_return, lambda: None)
        integral = avg * tf.cast(
            z.convert_to_tensor(space.rect_area()), dtype=avg.dtype
        )
        integrals.append(integral)
    return z.reduce_sum(integrals, axis=0)
    # return z.to_real(integral, dtype=dtype)


# TODO(Mayou36): Make more flexible for sampling
# @z.function
def normalization_nograd(
    func, n_axes, batch_size, num_batches, dtype, space, x=None, shape_after=()
):
    upper, lower = space.rect_limits
    lower = z.convert_to_tensor(lower, dtype=dtype)
    upper = z.convert_to_tensor(upper, dtype=dtype)

    def body(batch_num, mean):
        start_idx = batch_num * batch_size
        end_idx = start_idx + batch_size
        indices = tf.range(start_idx, end_idx, dtype=tf.int32)
        samples_normed = tfp.mcmc.sample_halton_sequence(
            n_axes,
            # num_results=batch_size,
            sequence_indices=indices,
            dtype=dtype,
            randomized=False,
        )
        # halton_sample = tf.random_uniform(shape=(n_axes, batch_size), dtype=dtype)
        samples_normed.set_shape((batch_size, n_axes))
        samples_normed = znp.expand_dims(samples_normed, axis=0)
        samples = samples_normed * (upper - lower) + lower
        func_vals = func(samples)
        if shape_after == ():
            reduce_axis = None
        else:
            reduce_axis = 1
            if len(func_vals.shape) == 1:
                func_vals = znp.expand_dims(func_vals, -1)
        batch_mean = znp.mean(func_vals, axis=reduce_axis)  # if there are gradients
        err_weight = 1 / tf.cast(batch_num + 1, dtype=tf.float64)

        do_print = False
        if do_print:
            tf.print(batch_num + 1)
        return batch_num + 1, mean + err_weight * (batch_mean - mean)

    cond = lambda batch_num, _: batch_num < num_batches

    initial_mean = tf.constant(0, shape=shape_after, dtype=dtype)
    initial_body_args = (0, initial_mean)
    _, final_mean = tf.while_loop(
        cond=cond,
        body=body,
        loop_vars=initial_body_args,
        parallel_iterations=1,
        swap_memory=False,
        back_prop=True,
    )
    # def normalization_grad(x):
    return final_mean


# @z.function
def normalization_chunked(
    func, n_axes, batch_size, num_batches, dtype, space, x=None, shape_after=()
):
    x_is_none = x is None

    @tf.custom_gradient
    def normalization_func(x):
        if x_is_none:
            x = None
        value = normalization_nograd(
            func=func,
            n_axes=n_axes,
            batch_size=batch_size,
            num_batches=num_batches,
            dtype=dtype,
            space=space,
            x=x,
            shape_after=shape_after,
        )

        def grad_fn(dy, variables=None):
            if variables is None:
                return dy, None
            with tf.GradientTape() as tape:
                value = normalization_nograd(
                    func=func,
                    n_axes=n_axes,
                    batch_size=batch_size,
                    num_batches=num_batches,
                    dtype=dtype,
                    space=space,
                    x=x,
                    shape_after=shape_after,
                )

            return dy, tape.gradient(value, variables)

        return value, grad_fn

    fake_x = 1 if x_is_none else x
    return normalization_func(fake_x)


# @z.function
def chunked_average(func, x, num_batches, batch_size, space, mc_sampler):
    lower, upper = space.limits

    fake_resource_var = tf.Variable(
        "fake_hack_ResVar_for_custom_gradient", initializer=z.constant(4242.0)
    )
    fake_x = z.constant(42.0) * fake_resource_var

    @tf.custom_gradient
    def dummy_func(fake_x):  # to make working with custom_gradient
        if x is not None:
            raise WorkInProgressError("partial not yet implemented")

        def body(batch_num, mean):
            if mc_sampler == tfp.mcmc.sample_halton_sequence:
                start_idx = batch_num * batch_size
                end_idx = start_idx + batch_size
                indices = tf.range(start_idx, end_idx, dtype=tf.int32)
                sample = mc_sampler(
                    space.n_obs,
                    sequence_indices=indices,
                    dtype=ztypes.float,
                    randomized=False,
                )
            else:
                sample = mc_sampler(shape=(batch_size, space.n_obs), dtype=ztypes.float)
            sample = tf.guarantee_const(sample)
            sample = (np.array(upper[0]) - np.array(lower[0])) * sample + lower[0]
            sample = znp.transpose(a=sample)
            sample = func(sample)
            sample = tf.guarantee_const(sample)

            batch_mean = znp.mean(sample)
            batch_mean = tf.guarantee_const(batch_mean)
            err_weight = 1 / tf.cast(batch_num + 1, dtype=tf.float64)
            # err_weight /= err_weight + 1
            # print_op = tf.print(batch_mean)
            do_print = False
            if do_print:
                tf.print(batch_num + 1, mean, err_weight * (batch_mean - mean))

            return batch_num + 1, mean + err_weight * (batch_mean - mean)

        cond = lambda batch_num, _: batch_num < num_batches

        initial_mean = tf.convert_to_tensor(value=0, dtype=ztypes.float)
        _, final_mean = tf.while_loop(
            cond=cond,
            body=body,
            loop_vars=(0, initial_mean),
            parallel_iterations=1,
            swap_memory=False,
            back_prop=False,
            maximum_iterations=num_batches,
        )

        def dummy_grad_with_var(dy, variables=None):
            raise WorkInProgressError("Who called me? Mayou36")
            if variables is None:
                raise WorkInProgressError(
                    "Is this needed? Why? It's not a NN. Please make an issue."
                )

            def dummy_grad_func(x):
                values = func(x)
                if variables:
                    gradients = tf.gradients(ys=values, xs=variables, grad_ys=dy)
                else:
                    gradients = None
                return gradients

            return chunked_average(
                func=dummy_grad_func,
                x=x,
                num_batches=num_batches,
                batch_size=batch_size,
                space=space,
                mc_sampler=mc_sampler,
            )

        def dummy_grad_without_var(dy):
            return dummy_grad_with_var(dy=dy, variables=None)

        do_print = False
        if do_print:
            tf.print("Total mean calculated = ", final_mean)

        return final_mean, dummy_grad_with_var

    try:
        return dummy_func(fake_x)
    except TypeError:
        return dummy_func(fake_x)


class PartialIntegralSampleData(BaseDimensional, ZfitData):
    def __init__(self, sample: list[tf.Tensor], space: ZfitSpace):
        """Takes a list of tensors and "fakes" a dataset. Useful for tensors with non-matching shapes.

        Args:
            sample:
            space:
        """
        if not isinstance(sample, list):
            raise TypeError("Sample has to be a list of tf.Tensors")
        super().__init__()
        self._space = space
        self._sample = sample
        self._reorder_indices_list = list(range(len(sample)))

    @property
    def weights(self):
        raise NotImplementedError(
            "Weights for PartialIntegralsampleData are not implemented. Are they needed?"
        )

    @property
    def space(self) -> zfit.Space:
        return self._space

    def sort_by_axes(self, axes, allow_superset: bool = True):
        axes = convert_to_container(axes)
        new_reorder_list = [
            self._reorder_indices_list[self.space.axes.index(ax)] for ax in axes
        ]
        value = self.space.with_axes(axes=axes), new_reorder_list

        getter = lambda: (self.space, self._reorder_indices_list)

        def setter(value):
            self._space, self._reorder_indices_list = value

        return TemporarilySet(value=value, getter=getter, setter=setter)

    def sort_by_obs(self, obs, allow_superset: bool = True):
        obs = convert_to_container(obs)
        new_reorder_list = [
            self._reorder_indices_list[self.space.obs.index(ob)] for ob in obs
        ]

        value = self.space.with_obs(obs=obs), new_reorder_list

        getter = lambda: (self.space, self._reorder_indices_list)

        def setter(value):
            self._space, self._reorder_indices_list = value

        return TemporarilySet(value=value, getter=getter, setter=setter)

    def value(self, obs: list[str] = None):
        return self

    def unstack_x(self, always_list=False):
        unstacked_x = [self._sample[i] for i in self._reorder_indices_list]
        if len(unstacked_x) == 1 and not always_list:
            unstacked_x = unstacked_x[0]
        return unstacked_x

    def __hash__(self) -> int:
        return id(self)


class AnalyticIntegral:
    def __init__(self, *args, **kwargs):
        """Hold analytic integrals and manage their dimensions, limits etc."""
        super().__init__(*args, **kwargs)
        self._integrals = collections.defaultdict(dict)

    def get_max_axes(
        self, limits: ztyping.LimitsType, axes: ztyping.AxesTypeInput = None
    ) -> tuple[int]:
        """Return the maximal available axes to integrate over analytically for given limits.

        Args:
            limits: The integral function will be able to integrate over this limits
            axes: The axes over which (or over a subset) it will integrate

        Returns:
            Tuple[int]:
        """
        if not isinstance(limits, ZfitSpace):
            raise TypeError("`limits` have to be a `ZfitSpace`")
        # limits = convert_to_space(limits=limits)

        return self._get_max_axes_limits(limits, out_of_axes=limits.axes)[
            0
        ]  # only axes

    def _get_max_axes_limits(
        self, limits, out_of_axes
    ):  # TODO: automatic caching? but most probably not relevant
        if out_of_axes:
            out_of_axes = frozenset(out_of_axes)
            implemented_axes = frozenset(
                d for d in self._integrals.keys() if d <= out_of_axes
            )
        else:
            implemented_axes = set(self._integrals.keys())
        implemented_axes = sorted(
            implemented_axes, key=len, reverse=True
        )  # iter through biggest first
        for axes in implemented_axes:
            limits_matched = [
                lim
                for lim, integ in self._integrals[axes].items()
                if integ.limits >= limits
            ]

            if limits_matched:  # one or more integrals available
                return tuple(sorted(axes)), limits_matched
        return (), ()  # no integral available for this axes

    def get_max_integral(
        self, limits: ztyping.LimitsType, axes: ztyping.AxesTypeInput = None
    ) -> None | Integral:
        """Return the integral over the ``limits`` with ``axes`` (or a subset of them).

        Args:
            limits:
            axes:

        Returns:
            Return a callable that integrated over the given limits.
        """
        limits = convert_to_space(limits=limits, axes=axes)

        axes, limits = self._get_max_axes_limits(limits=limits, out_of_axes=axes)
        axes = frozenset(axes)
        integrals = [self._integrals[axes][lim] for lim in limits]
        return max(integrals, key=lambda l: l.priority, default=None)

    def register(
        self,
        func: Callable,
        limits: ztyping.LimitsType,
        priority: int = 50,
        *,
        supports_norm: bool = False,
        supports_multiple_limits: bool = False,
    ) -> None:
        """Register an analytic integral.

        Args:
            func: The integral function. Takes 1 argument.
            axes: |dims_arg_descr|
            limits: |limits_arg_descr| ``Limits`` can be None if ``func`` works for any
            possible limits
            priority: If two or more integrals can integrate over certain limits, the one with the higher
                priority is taken (usually around 0-100).
            supports_norm: If True, norm_range will (if needed) be given to ``func`` as an argument.
            supports_multiple_limits: If True, multiple limits may be given as an argument to ``func``.
        """

        # if limits is False:
        #     raise ValueError("Limits for the analytical integral have to be specified or None (for any limits).")
        # if limits is None:
        #     limits = tuple((Space.ANY_LOWER, Space.ANY_UPPER) for _ in range(len(axes)))
        #     limits = convert_to_space(axes=axes, limits=limits)
        # else:
        #     limits = convert_to_space(axes=self.axes, limits=limits)
        # limits = limits.get_limits()
        if not isinstance(limits, ZfitSpace):
            raise TypeError("Limits for registering an integral have to be `ZfitSpace`")
        axes = frozenset(limits.axes)

        # add catching everything unsupported:
        func = supports(norm=supports_norm, multiple_limits=supports_multiple_limits)(
            func
        )
        limits = limits.with_axes(axes=tuple(sorted(limits.axes)))
        self._integrals[axes][limits] = Integral(
            func=func, limits=limits, priority=priority
        )  # TODO improve with
        # database-like access

    def integrate(
        self,
        x: ztyping.XType | None,
        limits: ztyping.LimitsType,
        axes: ztyping.AxesTypeInput = None,
        norm: ztyping.LimitsType = None,
        model: ZfitModel = None,
        params: dict = None,
    ) -> ztyping.XType:
        """Integrate analytically over the axes if available.

        Args:
            x: If a partial integration is made, x are the value to be evaluated for the partial
                integrated function. If a full integration is performed, this should be `None`.
            limits: The limits to integrate
            axes: The dimensions to integrate over
            norm: |norm_range_arg_descr|
            params: The parameters of the function


        Returns:
            Union[tf.Tensor, float]:

        Raises:
            AnalyticIntegralNotImplementedError: If the requested integral is not available.
        """
        if axes is None:
            axes = limits.axes
        axes = frozenset(axes)
        integral_holder = self._integrals.get(axes)
        # limits = convert_to_space(axes=self.axes, limits=limits)
        if integral_holder is None:
            raise AnalyticIntegralNotImplemented(
                f"Analytic integral is not available for axes {axes}"
            )
        integral_fn = self.get_max_integral(limits=limits)
        if integral_fn is None:
            raise AnalyticIntegralNotImplemented(
                f"Integral is available for axes {axes}, but not for limits {limits}"
            )

        with suppress(TypeError):
            return integral_fn(
                x=x, limits=limits, norm=norm, params=params, model=model
            )
        with suppress(TypeError):
            return integral_fn(limits=limits, norm=norm, params=params, model=model)

        with suppress(TypeError):
            return integral_fn(
                x=x, limits=limits, norm_range=norm, params=params, model=model
            )
        with suppress(TypeError):
            return integral_fn(
                limits=limits, norm_range=norm, params=params, model=model
            )
        assert False, "Could not integrate, unknown reason. Please fill a bug report."

        # with suppress(TypeError):
        #     return integral_fn(x=x, limits=limits, norm=norm, params=params, model=model)
        # with suppress(TypeError):
        #     return integral_fn(limits=limits, norm=norm, params=params, model=model)


class Integral:  # TODO analytic integral
    def __init__(self, func: Callable, limits: ZfitSpace, priority: int | float):
        """A lightweight holder for the integral function."""
        self.limits = limits
        self.integrate = func
        self.axes = limits.axes
        self.priority = priority

    def __call__(self, *args, **kwargs):
        return self.integrate(*args, **kwargs)


# to be "the" future integral class
class Integration:
    def __init__(self, mc_sampler, draws_per_dim, tol, max_draws, draws_simpson):
        self.tol = tol
        self.max_draws = max_draws
        self.mc_sampler = mc_sampler
        self.draws_per_dim = draws_per_dim
        self.draws_simpson = draws_simpson



#  Copyright (c) 2023 zfit

import tensorflow as tf

from ..settings import ztypes
from . import root_search


def find_practical_support_bandwidth(kernel, bandwidth, absolute_tolerance=10e-5):
    """Return the kernel support for practical purposes.

    Used to find a support value for computations for kernel functions without finite (bounded) support.
    """

    absolute_root_tolerance = 1e-3
    relative_root_tolerance = root_search.default_relative_root_tolerance(ztypes.float)
    function_tolerance = 0

    kernel_instance = kernel(loc=0, scale=bandwidth)

    def objective_fn(x):
        return kernel_instance.prob(x) - tf.constant(absolute_tolerance, ztypes.float)

    roots, value_at_roots, num_iterations, converged = root_search.brentq(
        objective_fn,
        tf.constant(0.0, dtype=ztypes.float),
        tf.constant(8.0, dtype=ztypes.float) * bandwidth,
        absolute_root_tolerance=absolute_root_tolerance,
        relative_root_tolerance=relative_root_tolerance,
        function_tolerance=function_tolerance,
    )

    return roots + absolute_root_tolerance


def convolve_1d_data_with_kernel(
    kernel, bandwidth, data, grid, support=None, fft_method="conv1d"
):
    kernel_grid_min = tf.math.reduce_min(grid)
    kernel_grid_max = tf.math.reduce_max(grid)

    num_grid_points = tf.size(grid, ztypes.int)
    num_intervals = num_grid_points - tf.constant(1, ztypes.int)
    space_width = kernel_grid_max - kernel_grid_min
    dx = space_width / tf.cast(num_intervals, ztypes.float)

    L = tf.cast(num_grid_points, ztypes.float)

    if support is not None:
        support_bandwidth = support * bandwidth
    else:
        support_bandwidth = find_practical_support_bandwidth(kernel, bandwidth)

    L = tf.minimum(tf.math.floor(support_bandwidth / dx), L)

    # Calculate the kernel weights
    zero = tf.constant(0, ztypes.float)
    kernel_grid = tf.linspace(
        zero, dx * L, tf.cast(L, ztypes.int) + tf.constant(1, ztypes.int)
    )
    kernel_weights = kernel(loc=zero, scale=bandwidth).prob(kernel_grid)
    kernel_weights = tf.concat(
        values=[tf.reverse(kernel_weights, axis=[0])[:-1], kernel_weights], axis=0
    )

    c = data
    k = kernel_weights

    if fft_method == "conv1d":
        c_size = tf.size(c, ztypes.int)
        c = tf.reshape(c, [1, c_size, 1], name="c")

        k_size = tf.size(k, ztypes.int)
        k = tf.reshape(k, [k_size, 1, 1], name="k")

        return tf.squeeze(tf.nn.conv1d(c, k, 1, "SAME"))

    else:
        P = tf.math.pow(
            tf.constant(2, ztypes.int),
            tf.cast(
                tf.math.ceil(
                    tf.math.log(
                        tf.constant(3.0, ztypes.float)
                        * tf.cast(num_grid_points, ztypes.float)
                        - tf.constant(1.0, ztypes.float)
                    )
                    / tf.math.log(tf.constant(2.0, ztypes.float))
                ),
                ztypes.int,
            ),
        )

        right_padding = (
            tf.cast(P, ztypes.int)
            - tf.constant(2, ztypes.int) * num_grid_points
            - tf.constant(1, ztypes.int)
        )
        left_padding = num_grid_points - tf.constant(1, ztypes.int)

        c = tf.pad(data, [[left_padding, right_padding]])
        k = tf.pad(kernel_weights, [[0, right_padding]])

        result = tf.signal.irfft(tf.signal.rfft(c) * tf.signal.rfft(k))
        start, end = tf.constant(2, ztypes.int) * num_grid_points - tf.constant(
            1, ztypes.int
        ), tf.constant(3, ztypes.int) * num_grid_points - tf.constant(1, ztypes.int)
        result = result[start:end]

        return result



#  Copyright (c) 2023 zfit

from __future__ import annotations

import typing
from collections.abc import Callable
from contextlib import suppress

import tensorflow_probability as tfp
import zfit_interface.typing as ztyping
from zfit_interface.pdf import ZfitPDF
from zfit_interface.variables import ZfitVar, ZfitSpace, ZfitParam

from zfit import convert_to_parameter, z
from zfit._variables.varsupport import VarSupports
from zfit.core.func import Func
from zfit.core.values import ValueHolder
from zfit.util.container import convert_to_container
from zfit.util.exception import (
    SpecificFunctionNotImplemented,
    NotExtendedPDFError,
    WorkInProgressError,
)


class Integration:
    _analytic_integrals = {}

    def __init__(self, mc_sampler=None, draws_per_dim=None, numeric_integrator=None):
        self._analytic_integrals = self._analytic_integrals.copy()
        if mc_sampler is None:
            mc_sampler = lambda *args, **kwargs: tfp.mcmc.sample_halton_sequence(
                *args, randomized=False, **kwargs
            )
        if numeric_integrator is None:
            numeric_integrator = False  # TODO
        if draws_per_dim is None:
            draws_per_dim = 40_000
        self.numeric_integrator = numeric_integrator
        self.mc_sampler = mc_sampler
        self.draws_per_dim = draws_per_dim

    def register_on_object(
        self, var: ztyping.Variable, func: Callable, overwrite: bool = False
    ):
        var = convert_to_container(var, frozenset)
        if var in self._analytic_integrals and not overwrite:
            raise ValueError(
                f"An analytic integral for {var} is already registered and 'overwrite' is "
                f"set to False."
            )
        self._analytic_integrals[var] = func

    def get_available(self, var):
        var = convert_to_container(var, frozenset)
        candidates = sorted(
            (v for v in self._analytic_integrals if var.issubset(v)), key=len
        )
        return {v: self._analytic_integrals[v] for v in candidates}

    @property
    def has_full(self, var):
        var = convert_to_container(var, frozenset)
        return len(list(self.get_available(var).keys()) + [[]][0]) == len(var)

    def has_partial(self, var):
        var = convert_to_container(var, frozenset)
        return bool(self.get_available(var))


class PDF(Func, ZfitPDF):
    def __init__(
        self,
        obs: typing.Mapping[str, ZfitSpace] = None,
        params: typing.Mapping[str, ZfitParam] = None,
        var: typing.Mapping[str, ZfitVar] = None,
        supports: typing.Mapping[str, typing.Mapping[str, VarSupports]] = None,
        extended: bool = None,
        norm: typing.Mapping[str, ZfitSpace] = None,
        label: str | None = None,
    ):
        self.supports = supports
        if norm is None:
            norm = obs.values()  # TODO: preprocess
        super().__init__(var=var, label=label)
        if norm is None:
            norm = self.space
        self.norm = norm
        if extended is not None:
            self._set_yield(extended)

        self.integration = Integration()

    def _set_yield(self, value):
        # if self.is_extended:
        #     raise AlreadyExtendedPDFError(f"Cannot extend {self}, is already extended.")
        value = convert_to_parameter(value)
        # self.add_cache_deps(value)  # TODO
        self._yield = value

    @property
    def is_extended(self) -> bool:
        """Flag to tell whether the model is extended or not.

        Returns:
            A boolean.
        """
        return self._yield is not None

    def __call__(self, var):
        if self.is_extended:
            return self.ext_pdf(var)
        else:
            return self.pdf(var)

    def _pdf(self, var, norm):
        raise SpecificFunctionNotImplemented

    def pdf(
        self,
        var: ztyping.VarInputType,
        norm: ztyping.NormInputType = None,
        *,
        options=None,
    ) -> ztyping.PDFReturnType:
        """Probability density function, normalized over ``norm``.

        Args:
          var: ``float`` or ``double`` ``Tensor``.
          norm: :py:class:`~zfit.Space` to normalize over

        Returns:
          :py:class:`tf.Tensor` of type `self.dtype`.
        """
        var = self._convert_check_input_var(var)
        norm = self._convert_check_input_norm(norm, var=var)
        if var.space is not None:
            return self.integrate(limits=var, norm=norm, options=options)
        value = self._call_pdf(var=var, norm=norm, options=options)
        return value
        # with self._convert_sort_x(var) as var:
        #     value = self._single_hook_pdf(x=var, norm_range=norm)
        #     if run.numeric_checks:
        #         z.check_numerics(value, message="Check if pdf output contains any NaNs of Infs")
        #     return z.to_real(value)

    @z.function(wraps="model")
    def _call_pdf(self, var, norm, *, options=None):
        return self._pdf(var, norm)  # TODO

    def _ext_pdf(self, var, norm):
        raise SpecificFunctionNotImplemented

    def ext_pdf(
        self,
        var: ztyping.VarInputType,
        norm: ztyping.NormInputType = None,
        *,
        options=None,
    ) -> ztyping.PDFReturnType:
        """Probability density function, normalized over ``norm``.OneDim.

        Args:
          var: ``float`` or ``double`` ``Tensor``.
          norm: :py:class:`~zfit.Space` to normalize over

        Returns:
          :py:class:`tf.Tensor` of type `self.dtype`.
        """
        if not self.is_extended:
            raise NotExtendedPDFError
        var = self._convert_check_input_var(var)
        norm = self._convert_check_input_norm(norm, var=var)
        if var.space is not None:
            return self.integrate(limits=var, norm=norm, options=options)
        return self._call_ext_pdf(var=var, norm=norm, options=options)

    @z.function(wraps="model")
    def _call_ext_pdf(self, var, norm, *, options=None):
        return self._ext_pdf(var, norm)  # TODO

    def _integrate(self, var, norm, options):
        raise SpecificFunctionNotImplemented

    def integrate(self, limits, norm=None, *, var=None, options=None):
        var = self._convert_check_input_var(limits, var)
        if var.space is None:
            raise ValueError(
                f"No space is given to integrate of {self}, needs at least one."
            )
        norm = self._convert_check_input_norm(norm, var=var)
        return self._call_integrate(var=var, norm=norm, options=options)

    @z.function(wraps="model")
    def _call_integrate(self, var, norm, options):
        with suppress(SpecificFunctionNotImplemented):
            return self._auto_integrate(var, norm, options=options)
        if self.is_extended:
            return (
                self._auto_ext_integrate(var, norm, options=options) / self.get_yield()
            )
        return self._fallback_integrate(var, norm, options=options)

    def _auto_integrate(self, var, norm, options):
        with suppress(SpecificFunctionNotImplemented):
            return self._integrate(var, norm, options=options)
        return self._fallback_integrate(var=var, norm=norm, options=options)

    def _fallback_integrate(self, var, norm, options):
        pass

    def _ext_integrate(self, var, norm, options):
        raise SpecificFunctionNotImplemented

    def _values(self, var=None, options=None):
        if self.is_extended:
            return self.rel_counts(var=var, options=options)
        else:
            return self.counts(var=var, options=options)

    def counts(self, *, var=None, norm=None, options=None):
        return self._call_counts(var=var, norm=norm, options=options)

    def _call_counts(self, var=None, norm=None, options=None):
        with suppress(SpecificFunctionNotImplemented):
            return self._counts(var, norm, options=options)  # TODO: auto_value?
        return self._call_ext_pdf(var=var, norm=norm, options=options)

    def _counts(self, var=None, norm=None, options=None):
        raise SpecificFunctionNotImplemented

    def rel_counts(self, *, var=None, norm=None, options=None):
        return self._call_rel_counts(var=var, norm=norm, options=options)

    def _call_rel_counts(self, var=None, norm=None, options=None):
        with suppress(SpecificFunctionNotImplemented):
            return self._rel_counts(var, norm, options=options)  # TODO: auto_value?
        return self._fallback_rel_counts(var=var, norm=norm, options=options)

    def _rel_counts(self, var=None, norm=None, options=None):
        raise SpecificFunctionNotImplemented

    def _fallback_rel_counts(self, var, norm, options):
        raise WorkInProgressError

    def ext_integrate(self, limits, norm=None, *, var=None, options=None):
        if not self.is_extended:
            raise NotExtendedPDFError
        var = self._convert_check_input_var(limits, var)
        if var.space is None:
            raise ValueError(
                f"No space is given to integrate of {self}, needs at least one."
            )
        norm = self._convert_check_input_norm(norm, var=var)
        return self._call_ext_integrate(var=var, norm=norm, options=options)

    @z.function(wraps="model")
    def _call_ext_integrate(self, var, norm, options):
        with suppress(SpecificFunctionNotImplemented):
            return self._auto_ext_integrate(var, norm, options=options)
        if self.is_extended:
            return self._auto_integrate(var, norm, options=options) * self.get_yield()
        return self._fallback_ext_integrate(var, norm, options=options)

    def _auto_ext_integrate(self, var, norm, options):
        return self._ext_integrate(var, norm, options=options)

    def _fallback_ext_integrate(self, var, norm, options):
        pass  # TODO
        # return self.integration.mixed(var, norm, options)

    def _convert_check_input_var(self, var):
        var = ValueHolder(var)
        return var  # TODO

    def _convert_check_input_norm(self, norm, var):
        if norm is None:
            norm = self.norm
        # return var  # TODO


class UnbinnedPDF(PDF):
    def __init__(
        self, obs, params=None, var=None, supports=None, extended=None, norm=None
    ):
        supports_default = "ext_pdf" if extended else "pdf"
        if supports is None:
            supports = {}
        if supports_default not in supports:
            supports[supports_default] = {}

        if obs is None:
            obs_supports = {}
        else:
            obs_supports = {
                axis: VarSupports(var=ob.name, data=True)
                for axis, ob in obs.items()
                if not isinstance(ob, VarSupports)
            }
        if params is None:
            params_supports = {}
        else:
            params_supports = {
                axis: VarSupports(var=p.name, scalar=True) for axis, p in params.items()
            }
        if var is None:
            var_supports = {}
        else:
            var_supports = var.copy()
        var_supports.update(obs_supports)
        var_supports.update(params_supports)
        if supports_default not in supports:
            supports[supports_default] = var_supports
        super().__init__(
            obs=obs,
            params=params,
            var=var,
            supports=supports,
            extended=extended,
            norm=norm,
        )


class HistPDF(PDF):
    def __init__(
        self,
        obs: typing.Mapping[str, ZfitSpace] = None,
        params: typing.Mapping[str, ZfitParam] = None,
        var: typing.Mapping[str, ZfitVar] = None,
        supports: typing.Mapping[str, typing.Mapping[str, VarSupports]] = None,
        extended: bool = None,
        norm: typing.Mapping[str, ZfitSpace] = None,
        label: str | None = None,
    ):
        supports_default = "counts" if extended else "rel_counts"
        if supports is None:
            supports = {}
        if supports_default not in supports:
            supports[supports_default] = {}

        if obs is None:
            obs_supports = {}
        else:
            obs_supports = {
                axis: VarSupports(var=ob.name, binned=True)
                for axis, ob in obs.items()
                if not isinstance(ob, VarSupports)
            }
        if params is None:
            params_supports = {}
        else:
            params_supports = {
                axis: VarSupports(var=p.name, scalar=True) for axis, p in params.items()
            }
        if var is None:
            var_supports = {}
        else:
            var_supports = var.copy()
        var_supports.update(obs_supports)
        var_supports.update(params_supports)
        supports[supports_default] = var_supports
        if "pdf" not in supports:
            supports["pdf"] = {
                axis: VarSupports(var=v.var, full=True)
                for axis, v in supports[supports_default].items()
            }
        if "ext_pdf" not in supports:
            supports["ext_pdf"] = {
                axis: VarSupports(var=v.var, full=True)
                for axis, v in supports[supports_default].items()
            }
        super().__init__(
            obs=obs,
            params=params,
            var=var,
            extended=extended,
            norm=norm,
            label=label,
            supports=supports,
        )

    def _ext_pdf(self, var, norm):  # TODO: normalization?
        counts = self._call_counts(var=var, norm=norm)
        binareas = var.binned.binning.areas
        densities = counts / binareas
        return densities

    def _pdf(self, var, norm):  # TODO: normalization?
        counts = self._call_rel_counts(var=var, norm=norm)
        binareas = var.binned.binning.areas
        densities = counts / binareas
        return densities



#  Copyright (c) 2023 zfit
from __future__ import annotations

import collections
import contextlib
import copy
import functools
from dataclasses import dataclass
from enum import Enum
from typing import (
    Any,
    Union,
    Mapping,
    Iterable,
    Dict,
    List,
    TypeVar,
    Optional,
    Tuple,
    Literal,
)

import numpy as np
import pydantic
import tensorflow as tf
from frozendict import frozendict
from pydantic import Field

from zfit.util.container import convert_to_container

try:
    from typing import Annotated
except ImportError:  # TODO(3.9): remove
    from typing_extensions import Annotated

from zfit.core.interfaces import (
    ZfitParameter,
    ZfitPDF,
    ZfitData,
    ZfitBinnedData,
    ZfitConstraint,
    ZfitLoss,
)
from zfit.core.serialmixin import ZfitSerializable
from zfit.util.warnings import warn_experimental_feature


@dataclass
class Aliases:
    hs3_type: str = "type"


alias1 = Aliases(hs3_type="type")


class Types:
    def __init__(self):
        """Hold all types that are used in the serialization, automatically collects parameters and PDFs.

        This holding of types in this way is needed to delay the evaluation of the annotations, which is necessary since
        pydantic will remember (statically) the types and not evaluate them again. Therefore, we delay the evaluation of
        the types until the first time the serializer is used by artificially blocking the forward references
        evaluation.
        """
        self._constraint_repr = []
        self._data_repr = []
        self._pdf_repr = []
        self._param_repr = []
        self.block_forward_refs = True
        self.alias = alias1
        self.DUMMYTYPE = TypeVar("DUMMYTYPE")

    def one_or_many(self, repr):
        """Returns either a single or a list of the given repr correctly annotated.

        Args:
            repr: The repr to be annotated.

        Returns:
            The annotated repr.
        """
        if self.block_forward_refs:
            raise NameError(
                "Internal error, should always be caught! If you see this, most likely the annotation"
                " evaluation was not postponed. To fix this, add a `from __future__ import annotations`"
                " and make sure to use Python 3.8+"
            )
        if len(repr) == 0:
            return None
        elif len(repr) == 1:
            return repr[0]
        else:
            return Union[
                Annotated[Union[tuple(repr)], Field(discriminator="hs3_type")],
                self.DUMMYTYPE,
            ]

    @property
    def PDFTypeDiscriminated(self):
        return self.one_or_many(self._pdf_repr)

    @property
    def DataTypeDiscriminated(self):
        return self.one_or_many(self._data_repr)

    @property
    def ConstraintTypeDiscriminated(self):
        return self.one_or_many(self._constraint_repr)

    @property
    def ParamTypeDiscriminated(self):
        return self.one_or_many(self._param_repr)

    @property
    def ListParamTypeDiscriminated(self):
        return List[self.ParamTypeDiscriminated]

    @property
    def ParamInputTypeDiscriminated(self):
        return Union[self.ParamTypeDiscriminated, float, int]

    @property
    def ListParamInputTypeDiscriminated(self):
        return List[self.ParamInputTypeDiscriminated]

    def register_repr(self, repr: Union[ZfitPDF, ZfitParameter]) -> None:
        """Register a repr to be used in the serialization such as PDF or Parameter.

        This is needed to make sure that objects which use any of these types can be recursively
        serialized. They are autamtically added to the correct type.
        For example, a `SumPDF` can be a sum of arbitrary PDFs, so it needs to be able to serialize
        any PDF. This is done by registering the PDFs to the serializer and assigning the datafield
        of `PDFTypeDiscriminated` to the `pdfs` field of `SumPDF`.

        Discriminated refers to the fact that they're not arbitrary types, but only the types that
        are registered and that exactly those will be used.

        Args:
            repr: The repr to be registered.
        """
        cls = repr._implementation

        if issubclass(cls, ZfitPDF):
            self._pdf_repr.append(repr)
        elif issubclass(cls, ZfitParameter):
            self._param_repr.append(repr)
        elif issubclass(cls, (ZfitData, ZfitBinnedData)):
            self._data_repr.append(repr)
        elif issubclass(cls, ZfitConstraint):
            self._constraint_repr.append(repr)


class SerializationTypeError(TypeError):
    pass


class Serializer:
    """Main serializer, to be used as a class only."""

    def __new__(cls, *args, **kwargs):
        raise TypeError(
            "Serializer should be used as a class, no instances are allowed"
        )

    types = Types()
    is_initialized = False

    constructor_repr = {}
    type_repr = {}
    _deserializing = False

    @classmethod
    def register(own_cls, repr: ZfitSerializable) -> None:
        """Register a repr to be used in the HS3 serialization.

        Args:
            repr: The repr to be registered.
        """
        cls = repr._implementation
        if not issubclass(cls, ZfitSerializable):
            raise TypeError(
                f"{cls} is not a subclass of ZfitSerializable. Possible solution: inherit from "
                f"the SerializableMixin"
            )

        if cls not in own_cls.constructor_repr:
            own_cls.constructor_repr[cls] = repr
        else:
            raise ValueError(f"Class {cls} already registered")

        hs3_type = repr.__fields__["hs3_type"].default
        cls.hs3_type = hs3_type
        cls.__annotations__["hs3_type"] = Literal[hs3_type]
        if hs3_type not in own_cls.type_repr:
            own_cls.type_repr[hs3_type] = repr
        else:
            raise ValueError(f"Type {hs3_type} already registered")

        own_cls.types.register_repr(repr)

    @classmethod
    def initialize(cls) -> None:
        """Initialize the serializer by evaluating all the forward references.

        This is a necessary implementation trick to work properly with pydantics caching as we cannot modify an existing
        Type after it has been created. This is necessary to register new types as possible options (say a new PDF is
        registered and can be a possibility in a SumPDF) on the fly.
        """
        if not cls.is_initialized:
            cls.types.block_forward_refs = False
            for repr in cls.constructor_repr.values():
                repr.update_forward_refs(
                    **{"Union": Union, "List": List, "Literal": Literal}
                )
            cls.is_initialized = True

    @classmethod
    @warn_experimental_feature
    def to_hs3(
        cls,
        obj: Union[
            Union[List[ZfitPDF], Tuple[ZfitPDF], ZfitPDF],
            Union[List[ZfitLoss], Tuple[ZfitLoss], ZfitLoss],
        ],
    ) -> Mapping[str, Any]:
        """Serialize a PDF or a list of PDFs to a JSON string according to the HS3 standard.

        .. warning::
            This is an experimental feature and the API might change in the future. DO NOT RELY ON THE OUTPUT FOR
            ANYTHING ELSE THAN TESTING.

        THIS FUNCTION DOESN'T YET ADHERE TO HS3 (but just as a proxy).

        |@doc:hs3.explain| HS3 is the `HEP Statistics Serialization Standard <https://github.com/hep-statistics-serialization-standard/hep-statistics-serialization-standard>`_.
                   It is a JSON/YAML-based serialization that is a
                   coordinated effort of the HEP community to standardize the serialization of statistical models. The standard
                   is still in development and is not yet finalized. This function is experimental and may change in the future. |@docend:hs3.explain|

        Args:
            obj: The PDF or list of PDFs to be serialized.

        Returns:
            mapping: The serialized objects as a mapping.
                     |@doc:hs3.layout.explain| The keys in the HS3 format are:
                    - 'distributions': list of PDFs
                    - 'variables': list of variables, i.e. ``zfit.Space`` and ``zfit.Parameter`` (or more generally parameters)
                    - 'loss': list of losses
                    - 'data': list of data
                    - 'metadata': contains the version of the HS3 format and the
                      zfit version used to create the file |@docend:hs3.layout.explain|
        """

        cls.initialize()

        serial_kwargs = {"exclude_none": True, "by_alias": True}
        # check if already HS3 format
        if isinstance(obj, collections.abc.Mapping):
            if (
                "distributions" in obj
                and isinstance(obj["distributions"], collections.abc.Mapping)
                and "variables" in obj
                and "metadata" in obj
            ):
                raise ValueError(
                    "Object seems to be already in HS3 format. If it contains PDFs, use `obj['distributions'].values()` instead of `obj` to get a valid conversion"
                )
            else:
                raise ValueError(
                    "Mappings are currently not supported. Use a PDF or a list of PDFs instead."
                )
        else:
            obj = convert_to_container(obj)
        from zfit.core.interfaces import ZfitPDF

        all_pdfs = all(isinstance(ob, ZfitPDF) for ob in obj)
        all_losses = all(isinstance(ob, ZfitLoss) for ob in obj)
        if not all_pdfs and not all_losses:
            raise TypeError("Only PDFs or losses can be serialized.")
        from zfit.core.serialmixin import ZfitSerializable

        if not all(isinstance(pdf, ZfitSerializable) for pdf in obj):
            raise SerializationTypeError("All distributions must be ZfitSerializable")
        import zfit

        out = {
            "metadata": {
                "HS3": {"version": "experimental"},
                "serializer": {"lib": "zfit", "version": zfit.__version__},
            },
            "distributions": {},
            "variables": {},
            "loss": {},
            "data": {},
            "constraints": {},
        }
        loss_number = range(len(obj))

        all_objs = {"data": [], "distributions": [], "constraints": [], "loss": []}
        if all_pdfs:
            all_objs["distributions"] = obj
        else:
            for loss in obj:
                all_objs["distributions"].extend(loss.model)
                all_objs["constraints"].extend(loss.constraints)
                all_objs["data"].extend(loss.data)
                all_objs["loss"].append(loss)
        all_objs = {key: set(val) for key, val in all_objs.items()}
        all_objs_cleaned = {key: {} for key in all_objs.keys()}
        # give all of the objects unique names
        for key, val in all_objs.items():
            for ob in val:
                name = ob.name
                if name in all_objs_cleaned[key]:
                    name = f"{name}_{iter(loss_number)}"
                all_objs_cleaned[key][name] = ob

        for name, pdf in all_objs_cleaned["distributions"].items():
            assert name not in out["distributions"], "Name should have been uniqueified"
            pdf_repr = pdf.get_repr().from_orm(pdf)
            out["distributions"][name] = pdf_repr.dict(**serial_kwargs)

            for param in pdf.get_params(
                floating=None, extract_independent=None
            ):  # TODO: this is not ideal, we should take the serialized params?
                if param.name not in out["variables"]:
                    paramdict = param.get_repr().from_orm(param).dict(**serial_kwargs)
                    del paramdict["type"]
                    out["variables"][param.name] = paramdict

            for ob in pdf.obs:
                if ob not in out["variables"]:
                    space = pdf.space.with_obs(ob)
                    spacedict = space.get_repr().from_orm(space).dict(**serial_kwargs)
                    del spacedict["type"]
                    out["variables"][ob] = spacedict

        for name, loss in all_objs_cleaned["loss"].items():
            out["loss"][name] = loss.get_repr().from_orm(loss).dict(**serial_kwargs)

        for name, data in all_objs_cleaned["data"].items():
            out["data"][name] = data.get_repr().from_orm(data).dict(**serial_kwargs)

        for name, constraint in all_objs_cleaned["constraints"].items():
            out["constraints"][name] = (
                constraint.get_repr().from_orm(constraint).dict(**serial_kwargs)
            )

        out = cls.post_serialize(out)

        return out

    @classmethod
    @warn_experimental_feature
    def from_hs3(
        cls, load: Mapping[str, Mapping]
    ) -> Mapping[str, Union[ZfitPDF, ZfitParameter]]:
        """Load a PDF or a list of PDFs from a JSON string according to the HS3 standard.

        .. warning::
            This is an experimental feature and the API might change in the future. DO NOT RELY ON THE OUTPUT FOR
            ANYTHING ELSE THAN TESTING.

        THIS FUNCTION DOESN'T YET ADHERE TO HS3 (but just as a proxy).

        |@doc:hs3.explain| HS3 is the `HEP Statistics Serialization Standard <https://github.com/hep-statistics-serialization-standard/hep-statistics-serialization-standard>`_.
                   It is a JSON/YAML-based serialization that is a
                   coordinated effort of the HEP community to standardize the serialization of statistical models. The standard
                   is still in development and is not yet finalized. This function is experimental and may change in the future. |@docend:hs3.explain|

        Args:
            load: The serialized objects as a mapping. |@doc:hs3.layout.explain| The keys in the HS3 format are:
                    - 'distributions': list of PDFs
                    - 'variables': list of variables, i.e. ``zfit.Space`` and ``zfit.Parameter`` (or more generally parameters)
                    - 'loss': list of losses
                    - 'data': list of data
                    - 'metadata': contains the version of the HS3 format and the
                      zfit version used to create the file |@docend:hs3.layout.explain|

        Returns:
            mapping: The PDFs and variables as a mapping to the original keys.
        """
        cls.initialize()
        # sanity checks, TODO
        if "variables" not in load:
            pass
        if "distributions" not in load:
            pass
        if "metadata" not in load:
            pass
        for param, paramdict in load["variables"].items():
            if "value" in paramdict:
                if paramdict.get("floating", True) is False:
                    paramdict["type"] = "ConstantParameter"
                else:
                    paramdict["type"] = "Parameter"
            elif "value_fn" in paramdict:
                paramdict["type"] = "ComposedParameter"
            else:
                paramdict["type"] = "Space"

        load = cls.pre_deserialize(load)

        out = {
            "variables": {},  # order matters! variables first
            "data": {},
            "constraints": {},
            "distributions": {},
            "loss": {},
        }
        assert (
            list(out)[0] == "variables"
        ), "Order changed, has to deserealize variables first"
        for kind, kindout in out.items():
            for name, obj_dict in load[kind].items():
                repr = Serializer.type_repr[obj_dict["type"]]
                repr_inst = repr(**obj_dict)
                kindout[name] = repr_inst.to_orm()
        out["metadata"] = load["metadata"].copy()

        out = cls.post_deserialize(out)
        return out

    @classmethod
    @contextlib.contextmanager
    def deserializing(cls):
        cls._deserializing = True
        yield
        cls._deserializing = False

    @classmethod
    def post_serialize(cls, out):
        # This is not very stable as it allows only one pass and cannot be applied multiple times (i.e. the replacement back of the params:
        # name is replaced by the dict, that's fine for *once*, but fails if done twice (as the "name" field will be replaced by the dict)
        for what in ["distributions", "loss", "data", "constraints"]:
            # replace constant parameters with their name
            const_params = frozendict(
                {"name": None, "type": "ConstantParameter", "floating": False}
            )
            replace_forward_const_param = {const_params: lambda x: x["name"]}
            out[what] = replace_matching(out[what], replace_forward_const_param)

            # replace composed parameters with their name
            composed_params = frozendict(
                {"name": None, "type": "ComposedParameter", "value_fn": None}
            )
            replace_forward_composed_param = {composed_params: lambda x: x["name"]}
            out[what] = replace_matching(out[what], replace_forward_composed_param)

            # replace parameters and spaces with their name
            parameter = frozendict({"name": None, "min": None, "max": None})
            replace_forward_param = {parameter: lambda x: x["name"]}
            out[what] = replace_matching(out[what], replace_forward_param)
        for parname, param in out["variables"].items():
            if "value_fn" in param:
                out["variables"][parname]["params"] = replace_matching(
                    out["variables"][parname]["params"], replace_forward_const_param
                )
                out["variables"][parname]["params"] = replace_matching(
                    out["variables"][parname]["params"], replace_forward_param
                )
        return out

    @classmethod
    def pre_deserialize(cls, out):
        out = copy.deepcopy(out)
        replace_backward = {
            k: lambda x=k: out["variables"][x] for k in out["variables"].keys()
        }
        for parname, param in out["variables"].items():
            if param["type"] == "ComposedParameter":
                param["params"] = replace_matching(param["params"], replace_backward)
        for what in ["distributions", "loss", "data", "constraints"]:
            out[what] = replace_matching(out[what], replace_backward)
        return out

    @classmethod
    def post_deserialize(cls, out):
        return out


TYPENAME = "hs3_type"


def elements_match(mapping, replace):
    found = False
    for match, replacement in replace.items():
        if isinstance(mapping, Mapping) and isinstance(match, Mapping):
            for k, v in match.items():
                if k not in mapping:
                    break
                if v is None:
                    continue  # fine so far, a "free field"
                val = mapping.get(k)

                if val == v:
                    continue  # also fine
                break  # we're not fine, so let's stop here
            else:
                found = True
                break
        try:
            direct_hit = bool(mapping == match)
        except ValueError as error:
            if (
                "The truth value of an array with more than one element is ambiguous"
                in str(error)
            ):
                direct_hit = bool((mapping == match).all())
            else:
                raise
        except TypeError:
            continue
        found = direct_hit or found
        if found:
            break
    else:
        return False, None
    return True, replacement(mapping)


def replace_matching(mapping, replace):
    # we need to test in the very beginning, it could be that the structure is a match
    is_match, new_map = elements_match(mapping, replace)
    if is_match:
        return new_map

    mapping = copy.copy(mapping)
    if isinstance(mapping, Mapping):
        for k, v in mapping.items():
            mapping[k] = replace_matching(v, replace)
    elif isinstance(mapping, Iterable) and not (
        isinstance(mapping, (str, (ZfitParameter, np.ndarray))) or tf.is_tensor(mapping)
    ):
        replaced_list = [replace_matching(v, replace) for v in mapping]
        mapping = type(mapping)(replaced_list)
    return mapping


def convert_to_orm(init):
    if isinstance(init, Mapping):
        for k, v in init.items():
            from zfit.core.interfaces import ZfitParameter, ZfitSpace

            from zfit.core.data import LightDataset

            if (
                not isinstance(v, (Iterable, Mapping))
                or isinstance(
                    v,
                    (
                        ZfitParameter,
                        ZfitSpace,
                        ZfitData,
                        ZfitBinnedData,
                        LightDataset,
                        np.ndarray,
                        str,
                        bytes,
                        bool,
                        int,
                        float,
                    ),
                )  # skip to not trigger the "in"
                or tf.is_tensor(v)
            ):
                continue
            elif TYPENAME in v:
                type_ = v[TYPENAME]
                init[k] = Serializer.type_repr[type_](**v).to_orm()
            else:
                init[k] = convert_to_orm(v)

        if TYPENAME in init:  # dicts can be the raw data we want.
            cls = Serializer.type_repr[init[TYPENAME]]
            return cls(**init).to_orm()

    elif isinstance(init, (list, tuple)):
        init = type(init)([convert_to_orm(v) for v in init])
    return init


def to_orm_init(func):
    @functools.wraps(func)
    def wrapper(self, init, **kwargs):
        init = convert_to_orm(init)
        return func(self, init, **kwargs)

    return wrapper


class MODES(Enum):
    orm = "orm"
    repr = "repr"


class BaseRepr(pydantic.BaseModel):
    _implementation = pydantic.PrivateAttr()
    _context = pydantic.PrivateAttr(None)
    _constructor = pydantic.PrivateAttr(None)

    dictionary: Optional[Dict] = Field(alias="dict")
    tags: Optional[List[str]] = None

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if (
            cls._implementation is not None
        ):  # TODO: better way to catch if constructor is set vs BaseClass?
            Serializer.register(cls)

    class Config:
        orm_mode = True
        arbitrary_types_allowed = True
        allow_population_by_field_name = True
        smart_union = True

    @classmethod
    def orm_mode(cls, v):
        if cls._context is None:
            raise ValueError("No context set!")
        return cls._context == MODES.orm

    @classmethod
    def from_orm(cls: pydantic.BaseModel, obj: Any) -> BaseRepr:
        old_mode = cls._context
        try:
            cls._context = MODES.orm
            out = super().from_orm(obj)
        finally:
            cls._context = old_mode
        return out

    def to_orm(self):
        old_mode = type(self)._context
        try:
            type(self)._context = MODES.repr
            if self._implementation is None:
                raise ValueError("No implementation registered!")
            init = self.dict(exclude_none=True)
            type_ = init.pop("hs3_type")
            # assert type_ == self.hs3_type
            out = self._to_orm(init)
        finally:
            type(self)._context = old_mode
        return out

    @to_orm_init
    def _to_orm(self, init):
        constructor = self._constructor
        if constructor is None:
            constructor = self._implementation
        return constructor(**init)



#  Copyright (c) 2024 zfit

from ._data.binneddatav1 import BinnedData
from ._variables.axis import RegularBinning, VariableBinning
from .core.data import Data

__all__ = ["Data", "BinnedData", "RegularBinning", "VariableBinning"]



#  Copyright (c) 2024 zfit

from __future__ import annotations

import collections
import contextlib
import functools
import math as _mt
import warnings
from collections import defaultdict
from collections.abc import Callable
from typing import Any
from weakref import WeakSet

import numpy as np
import tensorflow as tf

import zfit.z.numpy as znp
from ..settings import ztypes
from ..util.exception import BreakingAPIChangeError
from ..util.warnings import warn_advanced_feature


def constant(value, dtype=ztypes.float, shape=None, name="Const", verify_shape=None):
    return tf.constant(value, dtype=dtype, shape=shape, name=name)


pi = np.float64(_mt.pi)


def to_complex(number, dtype=ztypes.complex):
    return tf.cast(number, dtype=dtype)


def to_real(x, dtype=ztypes.float):
    return znp.asarray(tf.cast(x, dtype=dtype))


def abs_square(x):
    return znp.real(x * znp.conj(x))


def nth_pow(x, n):
    """Calculate the nth power of the complex Tensor x.

    Args:
        x:
        n: Power of x, has to be a positive int
        name: No effect, for API compatibility with tf.pow
    """
    if n < 0:
        raise ValueError(f"n (power) has to be >= 0. Currently, n={n}")

    power = to_complex(1.0)
    for _ in range(n):
        power *= x
    return power


def unstack_x(
    value: Any,
    num: Any = None,
    axis: int = -1,
    always_list: bool = False,
    name: str = "unstack_x",
):
    """Unstack a Data object and return a list of (or a single) tensors in the right order.

    Args:
        value:
        num:
        axis:
        always_list: If True, also return a list if only one element.
        name:

    Returns:
        Union[List[tensorflow.python.framework.ops.Tensor], tensorflow.python.framework.ops.Tensor, None]:
    """
    if isinstance(value, list):
        if len(value) == 1 and not always_list:
            value = value[0]
        return value
    try:
        return value.unstack_x(always_list=always_list)
    except AttributeError:
        unstacked_x = tf.unstack(value=value, num=num, axis=axis, name=name)
    if len(unstacked_x) == 1 and not always_list:
        assert isinstance(unstacked_x, list), (
            "unstacked_x has to be a list, otherwise this is a bug. Please report on github: "
            "https://github.com/zfit/zfit/issues/new?assignees=&labels=bug&projects=&template="
            "bug_report.md&title=[ASSERT]%20unstack_x%20does%20not%20provide%20a%20list,%20internal%20error"
        )
        unstacked_x = unstacked_x[0]
    return unstacked_x


def stack_x(values, axis: int = -1, name: str = "stack_x"):
    return tf.stack(values=values, axis=axis, name=name)


# random sampling


def convert_to_tensor(value, dtype=None, name=None, preferred_dtype=None):
    return tf.convert_to_tensor(
        value=value, dtype=dtype, name=name, dtype_hint=preferred_dtype
    )


def safe_where(
    condition: tf.Tensor,
    func: Callable,
    safe_func: Callable,
    values: tf.Tensor,
    value_safer: Callable = tf.ones_like,
) -> tf.Tensor:
    """Like :py:func:`tf.where` but fixes gradient `NaN` if func produces `NaN` with certain `values`.

    Args:
        condition: Same argument as to :py:func:`tf.where`, a boolean :py:class:`tf.Tensor`
        func: Function taking `values` as argument and returning the tensor _in case
            condition is True_. Equivalent `x` of :py:func:`tf.where` but as function.
        safe_func: Function taking `values` as argument and returning the tensor
            _in case the condition is False_, Equivalent `y` of :py:func:`tf.where` but as function.
        values: Values to be evaluated either by `func` or `safe_func` depending on
            `condition`.
        value_safer: Function taking `values` as arguments and returns "safe" values
            that won't cause troubles when given to`func` or by taking the gradient with respect
            to `func(value_safer(values))`.

    Returns:
        :py:class:`tf.Tensor`:
    """
    safe_x = tf.where(condition=condition, x=values, y=value_safer(values))
    result = tf.where(condition=condition, x=func(safe_x), y=safe_func(values))
    return result


def run_no_nan(func, x):
    from zfit.core.data import Data

    value_with_nans = func(x=x)
    if value_with_nans.dtype in (tf.complex128, tf.complex64):
        value_with_nans = znp.real(value_with_nans) + znp.imag(
            value_with_nans
        )  # we care only about NaN or not
    finite_bools = znp.isfinite(tf.cast(value_with_nans, dtype=tf.float64))
    finite_indices = tf.where(finite_bools)
    new_x = tf.gather_nd(params=x, indices=finite_indices)
    new_x = Data.from_tensor(obs=x.obs, tensor=new_x)
    vals_no_nan = func(x=new_x)
    result = tf.scatter_nd(
        indices=finite_indices,
        updates=vals_no_nan,
        shape=tf.shape(input=value_with_nans, out_type=finite_indices.dtype),
    )
    return result


class FunctionWrapperRegistry:
    registries = WeakSet()
    allow_jit = True
    DEFAULT_CACHE_SIZE = 40
    _DEFAULT_DO_JIT_TYPES = defaultdict(lambda: True)
    _DEFAULT_DO_JIT_TYPES.update(
        {
            None: True,
            "model": False,
            "loss": True,
            "sample": True,
            "model_sampling": True,
            "zfit_tensor": True,
            "tensor": True,
        }
    )

    do_jit_types = _DEFAULT_DO_JIT_TYPES.copy()

    def __init__(
        self,
        wraps=None,
        stateless_args=None,
        cachesize=None,
        keepalive=None,
        **kwargs_user,
    ) -> None:
        """`tf.function`-like decorator with additional cache-invalidation functionality.

        Args:
            keepalive:
            **kwargs_user: arguments to `tf.function`
        """
        super().__init__()
        if cachesize is None:
            cachesize = self.DEFAULT_CACHE_SIZE
        if stateless_args is None:
            stateless_args = False
        if keepalive is None:
            keepalive = False
        self._initial_user_kwargs = kwargs_user
        self._deleted_cachers = collections.Counter()

        self.registries.add(self)  # TODO: remove?
        self.python_func = None
        self.wrapped_func = None

        if wraps not in self.do_jit_types:
            from ..settings import run

            self.do_jit_types[wraps] = bool(run.get_graph_mode())
        self.wraps = wraps
        self.stateless_args = stateless_args

        self.function_cache = collections.deque()
        self.reset(**self._initial_user_kwargs)
        self.currently_traced = set()
        self.cachesize = cachesize
        self.keepalive = keepalive

    @property
    def do_jit(self):
        return self.do_jit_types[self.wraps] and self.allow_jit

    def reset(self, **kwargs_user):
        kwargs = dict(autograph=False, reduce_retracing=False)
        kwargs.update(self._initial_user_kwargs)
        kwargs.update(kwargs_user)
        self.tf_function_kwargs = kwargs
        self.function_cache.clear()

    def set_graph_cache_size(self, cachesize: int | None = None):
        """Set the size of the graph cache.

        Args:
            cachesize: Size of the cache. If None, the default size is used.
        """
        if cachesize is None:
            cachesize = self.DEFAULT_CACHE_SIZE
        self.cachesize = cachesize
        while len(self.function_cache) >= self.cachesize:
            self.function_cache.popleft()

    @property
    def tf_function(self):
        function = tf.function(**self.tf_function_kwargs)

        return function

    def __call__(self, func):
        keepalive = self.keepalive
        wrapped_func = self.tf_function(func)
        cache = self.function_cache
        deleted_cachers = self._deleted_cachers
        from ..util.cache import FunctionCacheHolder

        def concrete_func(*args, **kwargs):
            if not self.do_jit or func in self.currently_traced:
                return func(*args, **kwargs)

            self.currently_traced.add(func)
            nonlocal wrapped_func

            def deleter(proxy):
                with contextlib.suppress(ValueError):
                    cache.remove(function_holder)

            function_holder = FunctionCacheHolder(
                func,
                wrapped_func,
                args,
                kwargs,
                deleter=deleter,
                stateless_args=self.stateless_args,
            )
            #
            try:
                func_holder_index = cache.index(function_holder)
            except ValueError:
                wrapped_func = self.tf_function(func)
                func_to_run = wrapped_func
                function_holder = FunctionCacheHolder(
                    func,
                    wrapped_func,
                    args,
                    kwargs,
                    deleter=deleter,
                    stateless_args=self.stateless_args,
                    keepalive=keepalive,
                )
                if len(cache) >= self.cachesize:
                    popped_holder = cache.popleft()
                    hash_popped_holder = hash(popped_holder)
                    deleted_cachers.update((hash_popped_holder,))
                    if self._deleted_cachers[hash_popped_holder] > 3:
                        warnings.warn(
                            f"Function {function_holder.python_func} was removed from the cache more than 3"
                            f" times (and getting recompiled). Maybe consider increasing the cache size"
                            f" using `zfit.run.set_graph_cache_size(...)`, the current size is {self.cachesize}."
                        )

                        self._deleted_cachers - collections.Counter(
                            {hash(function_holder): int(-1e100)}
                        )  # won't be warned again
                    del popped_holder
                cache.append(function_holder)
            else:
                func_to_run = cache[func_holder_index].wrapped_func

            try:
                result = func_to_run(*args, **kwargs)
            finally:
                self.currently_traced.remove(func)
            return result

        return concrete_func


# equivalent to tf.function
def function(func=None, *, stateless_args=None, cachesize=None, **kwargs):
    """JIT/Graph compilation of functions, `tf.function`-like with additional cache-invalidation functionality.

    Args:
        func: Function to be compiled.
        stateless_args: If True, the function is assumed to be stateless and *does not depend on the name of tf.Variables*
            but only needs the values of the variables. This is not the case for taking gradients, for example.
        cachesize: Size of the cache. If None, the default size is used.
        **kwargs: arguments to `tf.function`

    Returns:
    """
    if stateless_args is None:
        stateless_args = False
    if callable(func):
        wrapper = FunctionWrapperRegistry()
        return wrapper(func)
    elif func:
        raise ValueError(
            "All argument have to be key-word only. `func` must not be used"
        )
    else:
        return FunctionWrapperRegistry(**kwargs, stateless_args=stateless_args)


# legacy, remove 0.6
def function_tf_input(*_, **__):
    raise BreakingAPIChangeError(
        "This function has been removed. Use `z.function(wraps='zfit_tensor') or your"
        "own category"
    )


# legacy, remove 0.6
def function_sampling(*_, **__):
    raise BreakingAPIChangeError(
        "This function has been removed. Use `z.function(wraps='zfit_sampling') or your"
        "own category"
    )


@functools.wraps(tf.py_function)
def py_function(func, inp, Tout, name=None):
    from .. import settings

    if not settings.options["numerical_grad"]:
        warn_advanced_feature(
            "Using py_function without numerical gradients. If the Python code does not contain any"
            " parametrization by `zfit.Parameter` or similar, this can work out. Otherwise, in case"
            " it depends on those, you may want to set `zfit.run.set_autograd_mode(=False)`.",
            identifier="py_func_autograd",
        )

    return tf.py_function(func=func, inp=inp, Tout=Tout, name=name)



#  Copyright (c) 2022 zfit

from . import (
    basefunc,
    basemodel,
    baseobject,
    basepdf,
    data,
    integration,
    interfaces,
    loss,
    operations,
    parameter,
    sample,
    space,
)



#  Copyright (c) 2024 zfit

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import zfit

import operator
from collections.abc import Iterable, Callable

from contextlib import suppress
from functools import reduce
from uhi.typing.plottable import PlottableHistogram

import numpy as np
import tensorflow as tf
import tensorflow_probability as tfp

import zfit
import zfit.z.numpy as znp
from zfit import z
from zfit._data.binneddatav1 import BinnedData, move_axis_obs, BinnedSampler
from .baseobject import BaseNumeric, extract_filter_params
from .binning import unbinned_to_binindex
from .data import Data
from .dimension import BaseDimensional
from .interfaces import (
    ZfitBinnedPDF,
    ZfitParameter,
    ZfitSpace,
    ZfitPDF,
    ZfitBinnedData,
    ZfitUnbinnedData,
    ZfitBinning,
)
from .parameter import convert_to_parameter
from .space import supports, convert_to_space
from .tensorlike import OverloadableMixinValues
from ..util import ztyping
from ..util.cache import GraphCachable
from ..util.container import convert_to_container
from ..util.deprecation import deprecated, deprecated_norm_range
from ..util.exception import (
    AlreadyExtendedPDFError,
    NotExtendedPDFError,
    SpaceIncompatibleError,
    SpecificFunctionNotImplemented,
    WorkInProgressError,
    NormNotImplemented,
    MultipleLimitsNotImplemented,
    BasePDFSubclassingError,
)

_BaseModel_USER_IMPL_METHODS_TO_CHECK = {}


def _BinnedPDF_register_check_support(has_support: bool):
    """Marks a method that the subclass either *has* to or *can't* use the ``@supports`` decorator.

    Args:
        has_support (bool): If True, flags that it **requires** the ``@supports`` decorator. If False,
            flags that the ``@supports`` decorator is **not allowed**.
    """
    if not isinstance(has_support, bool):
        raise TypeError("Has to be boolean.")

    def register(func):
        """Register a method to be checked to (if True) *has* ``support`` or (if False) has *no* ``support``.

        Args:
            func (function):
        Returns:
            function:
        """
        name = func.__name__
        _BaseModel_USER_IMPL_METHODS_TO_CHECK[name] = has_support
        func.__wrapped__ = _BinnedPDF_register_check_support
        return func

    return register


class BaseBinnedPDFV1(
    BaseNumeric,
    GraphCachable,
    BaseDimensional,
    OverloadableMixinValues,
    ZfitBinnedPDF,
):
    def __init__(self, obs, extended=None, norm=None, name=None, **kwargs):
        super().__init__(dtype=znp.float64, name=name, **kwargs)
        # self._name = name  # TODO: why is this needed?

        self._space = self._check_convert_obs_init(obs)
        self._yield = None
        self._norm = self._check_convert_norm_init(norm)
        if extended is None:
            extended = False
        if extended is not False:
            self._set_yield(extended)

    def _check_convert_obs_init(self, obs):
        if not isinstance(obs, ZfitSpace) or not obs.is_binned:
            raise ValueError(f"`obs` have to be a Space with binning, not {obs}.")
        return obs

    def _check_convert_norm_init(self, norm):
        if norm is not None:
            if not isinstance(norm, ZfitSpace) or not norm.has_limits:
                raise ValueError(
                    f"`norm` has to be None or a Space with limits, not {norm}."
                )
        return norm

    @classmethod
    def _subclass_check_support(cls, methods_to_check, wrapper_not_overwritten):
        for method_name, has_support in methods_to_check.items():
            method = getattr(cls, method_name)
            if hasattr(method, "__wrapped__"):
                if method.__wrapped__ == wrapper_not_overwritten:
                    continue  # not overwritten, fine

            # here means: overwritten
            if hasattr(method, "__wrapped__"):
                if method.__wrapped__ == supports:
                    if has_support:
                        continue  # needs support, has been wrapped
                    else:
                        raise BasePDFSubclassingError(
                            "Method {} has been wrapped with supports "
                            "but is not allowed to. Has to handle all "
                            "arguments.".format(method_name)
                        )
                elif has_support:
                    raise BasePDFSubclassingError(
                        "Method {} has been overwritten and *has to* be "
                        "wrapped by `supports` decorator (don't forget () )"
                        "to call the decorator as it takes arguments"
                        "".format(method_name)
                    )
                elif not has_support:
                    continue  # no support, has not been wrapped with
            else:
                if not has_support:
                    continue  # not wrapped, no support, need no

            # if we reach this points, somethings was implemented wrongly
            raise BasePDFSubclassingError(
                "Method {} has not been correctly wrapped with @supports "
                "OR has been wrapped but it should not be".format(method_name)
            )

    @property
    def axes(self):
        return self.space.binning

    def to_binneddata(self, **kwargs) -> zfit.data.BinnedData:
        """Create an Asimov dataset as ``BinnedData`` using either ``counts`` (for extended) or ``rel_counts``.

        Args:
            **kwargs (): arguments to ``counts`` or ``rel_counts``.

        Returns:
            BinnedData: Binned data representing the Asimov dataset of this PDF.
        """
        values = self.values(**kwargs)
        data = BinnedData.from_tensor(space=self.space, values=values)
        return data

    def to_hist(self, **kwargs):
        """Create an Asimov histogram as ``Hist`` using either ``counts`` (for extended) or ``rel_counts``.

        Args:
            **kwargs (): arguments to ``counts`` or ``rel_counts``.

        Returns:
            ``hist.Hist``: Histogram representing the Asimov dataset of this PDF.
        """
        return self.to_binneddata(**kwargs).to_hist()

    @property
    def space(self):
        return self._space

    def _set_yield(self, value):
        if self.is_extended:
            raise AlreadyExtendedPDFError(f"Cannot extend {self}, is already extended.")
        value = convert_to_parameter(value)
        self.add_cache_deps(value)
        self._yield = value

    def _get_dependencies(self) -> ztyping.DependentsType:
        return super()._get_dependencies()

    def _get_params(
        self,
        floating: bool | None = True,
        is_yield: bool | None = None,
        extract_independent: bool | None = True,
    ) -> set[ZfitParameter]:
        params = super()._get_params(
            floating, is_yield=is_yield, extract_independent=extract_independent
        )

        if is_yield is not False:
            if self.is_extended:
                yield_params = extract_filter_params(
                    self.get_yield(),
                    floating=floating,
                    extract_independent=extract_independent,
                )
                yield_params.update(params)  # putting the yields at the beginning
                params = yield_params
            elif is_yield is True:
                raise NotExtendedPDFError(
                    "PDF is not extended but only yield parameters were requested."
                )
        return params

    def _convert_input_binned_x(self, x, none_is_space=None):
        if x is None and none_is_space:
            return self.space
        if isinstance(x, PlottableHistogram) and not isinstance(x, ZfitBinnedData):
            x = BinnedData.from_hist(x)
        if not isinstance(x, ZfitBinnedData):
            if not isinstance(x, ZfitSpace):
                if not isinstance(x, ZfitUnbinnedData):
                    try:
                        x = Data.from_tensor(obs=self.obs, tensor=x)
                    except Exception as error:
                        raise TypeError(
                            f"Data to {self} has to be Binned Data, not {x}. (It can also be unbinned Data)"
                            + " but conversion to it failed (see also above) with the following error:"
                            + f" {error})"
                        ) from error

            # TODO: should we allow spaces? Or what?
        return x

    def _check_convert_norm(self, norm, none_is_error=False):
        if norm is None:
            norm = self.norm
        if norm is None:
            if none_is_error:
                raise ValueError("norm cannot be None for this function.")
        elif norm is not False and not isinstance(norm, ZfitSpace):
            raise TypeError(f"`norm` needs to be a binned ZfitSpace, not {norm}.")
        elif norm is not False and not norm.is_binned:
            norm = norm.with_binning(self.space.binning)
        return norm

    def _check_convert_limits(self, limits):
        if limits is None:
            limits = self.space
        if not isinstance(limits, ZfitSpace):
            limits = convert_to_space(obs=self.obs, limits=limits)
        if isinstance(limits, ZfitSpace) and not limits.is_binned:
            limits = limits.with_binning(self.space.binning)
        return limits

    @_BinnedPDF_register_check_support(True)
    def _pdf(self, x, norm):
        return self._call_rel_counts(x, norm=norm) / np.prod(
            self.space.binning.widths, axis=0
        )

    @deprecated_norm_range
    def pdf(
        self, x: ztyping.XType, norm: ztyping.LimitsType = None, *, norm_range=None
    ) -> ztyping.XType:
        """Probability density function, evaluated at ``x`` or in the bins of ``x``

        Args:
            x: |@doc:binnedpdf.pdf.x| Values to evaluate the PDF at.
               If this is a ``ZfitBinnedData``-like object, a histogram of *densities*
               will be returned. If x is a ``ZfitUnbinnedData``-like object, the densities will be
               evaluated at the points of ``x``. |@docend:binnedpdf.pdf.x|
            norm: |@doc:pdf.pdf.norm| Normalization of the function.
               By default, this is the ``norm`` of the PDF (which by default is the same as
               the space of the PDF). |@docend:pdf.pdf.norm|

        Returns:
            ``Array-like``: probability density
        """
        assert norm_range is None
        del norm_range  # taken care of in the deprecation decorator

        # convert the input argument to a standardized form
        x = self._convert_input_binned_x(x, none_is_space=True)
        norm = self._check_convert_norm(norm, none_is_error=True)

        # sort it and remember the original sorting
        original_space = x if isinstance(x, ZfitSpace) else x.space
        x = x.with_obs(self.space)

        # if it is unbinned, we get the binned version and gather the corresponding values
        is_unbinned = isinstance(x, ZfitUnbinnedData)
        binindices = None
        if is_unbinned:
            binindices = unbinned_to_binindex(x, self.space, flow=True)
            x = self.space

        values = self._call_pdf(x, norm=norm)

        if (
            binindices is not None
        ):  # because we have the flow, so we need to make it here with pads
            padded_values = znp.pad(
                values,
                znp.ones((z._get_ndims(values), 2), dtype=znp.float64),
                mode="constant",
            )  # for overflow
            ordered_values = tf.gather_nd(padded_values, indices=binindices)
        else:
            ordered_values = move_axis_obs(self.space, original_space, values)
        return znp.asarray(ordered_values)

    @z.function(wraps="model_binned")
    def _call_pdf(self, x, norm):
        with suppress(SpecificFunctionNotImplemented):
            return self._auto_pdf(x, norm)
        with suppress(SpecificFunctionNotImplemented):
            return self._auto_rel_counts(x, norm=norm) / np.prod(
                self.space.binning.widths, axis=0
            )
        return self._fallback_pdf(x, norm=norm)

    def _auto_pdf(self, x, norm):
        try:
            pdf = self._pdf(x, norm=norm)
        except NormNotImplemented:
            unnormed_pdf = self._pdf(x, norm=False)
            normalization = self.normalization(norm)
            pdf = unnormed_pdf / normalization
        return pdf

    def _fallback_pdf(self, x, norm):
        values = self._call_unnormalized_pdf(x)
        if norm is not False:
            values = values / self.normalization(norm)
        return values

    def _unnormalized_pdf(self, x):
        raise SpecificFunctionNotImplemented

    def _call_unnormalized_pdf(self, x):
        return self._unnormalized_pdf(x)

    @_BinnedPDF_register_check_support(True)
    def _ext_pdf(self, x, norm):
        raise SpecificFunctionNotImplemented

    @deprecated_norm_range
    def ext_pdf(
        self, x: ztyping.XType, norm: ztyping.LimitsType = None, *, norm_range=None
    ) -> ztyping.XType:
        """Probability density function scaled by yield, evaluated at ``x`` or in the bins of ``x``

        Args:
            x: |@doc:binnedpdf.pdf.x| Values to evaluate the PDF at.
               If this is a ``ZfitBinnedData``-like object, a histogram of *densities*
               will be returned. If x is a ``ZfitUnbinnedData``-like object, the densities will be
               evaluated at the points of ``x``. |@docend:binnedpdf.pdf.x|
            norm: |@doc:pdf.pdf.norm| Normalization of the function.
               By default, this is the ``norm`` of the PDF (which by default is the same as
               the space of the PDF). |@docend:pdf.pdf.norm|

        Returns:
            |@doc:binnedpdf.out.problike| If the input was unbinned, it returns an array
               of shape (nevents,). If the input was binned, the dimensions and ordering of
               the axes corresponds to the input axes. |@docend:binnedpdf.out.problike|
        """
        del norm_range  # should be taken care of by deprecation decorator
        if not self.is_extended:
            raise NotExtendedPDFError
        # convert the input argument to a standardized form
        x = self._convert_input_binned_x(x, none_is_space=True)
        norm = self._check_convert_norm(norm, none_is_error=True)
        # sort it and remember the original sorting
        original_space = x if isinstance(x, ZfitSpace) else x.space
        x = x.with_obs(self.space)

        # if it is unbinned, we get the binned version and gather the corresponding values
        is_unbinned = isinstance(x, ZfitUnbinnedData)
        binindices = None
        if is_unbinned:
            binindices = unbinned_to_binindex(x, self.space, flow=True)
            x = self.space

        values = self._call_ext_pdf(x, norm=norm)

        if (
            binindices is not None
        ):  # because we have the flow, so we need to make it here with pads
            padded_values = znp.pad(
                values,
                znp.ones((z._get_ndims(values), 2), dtype=znp.float64),
                mode="constant",
            )  # for overflow
            ordered_values = tf.gather_nd(padded_values, indices=binindices)
        else:
            ordered_values = move_axis_obs(self.space, original_space, values)
        return znp.asarray(ordered_values)

    @z.function(wraps="model_binned")
    def _call_ext_pdf(self, x, norm):
        with suppress(SpecificFunctionNotImplemented):
            return self._auto_ext_pdf(x, norm)
        with suppress(SpecificFunctionNotImplemented):
            return self._auto_counts(x, norm=norm) / np.prod(
                self.space.binning.widths, axis=0
            )
        return self._fallback_ext_pdf(x, norm=norm)

    def _auto_ext_pdf(self, x, norm):
        try:
            pdf = self._ext_pdf(x, norm=norm)
        except NormNotImplemented:
            unnormed_pdf = self._ext_pdf(x, norm=False)
            normalization = self.ext_normalization(norm)
            pdf = unnormed_pdf / normalization
        return pdf

    def _fallback_ext_pdf(self, x, norm):
        values = self._call_pdf(x, norm=norm)
        return values * self.get_yield()

    def normalization(self, limits, *, options=None) -> ztyping.NumericalTypeReturn:
        if limits is not None:
            norm = limits
        if options is None:
            options = {}
        norm = self._check_convert_norm(norm, none_is_error=True)
        return self._call_normalization(norm, options=options)

    def _call_normalization(self, norm, *, options):
        with suppress(SpecificFunctionNotImplemented):
            return self._normalization(norm, options=options)

        # fallback
        return self._call_integrate(norm, norm=False, options=None)

    @_BinnedPDF_register_check_support(True)
    def _normalization(self, limits, *, options):
        raise SpecificFunctionNotImplemented

    @_BinnedPDF_register_check_support(True)
    def _integrate(self, limits, norm, options):
        raise SpecificFunctionNotImplemented

    def integrate(
        self,
        limits: ztyping.LimitsType,
        norm: ztyping.LimitsType = None,
        *,
        options=None,
    ) -> ztyping.XType:
        if options is None:
            options = {}
        norm = self._check_convert_norm(norm)
        limits = self._check_convert_limits(limits)
        return self._call_integrate(limits, norm, options)

    @z.function(wraps="model_binned")
    def _call_integrate(self, limits, norm, options=None):
        if options is None:
            options = {}
        with suppress(SpecificFunctionNotImplemented):
            return self._auto_integrate(limits, norm, options)
        return self._fallback_integrate(limits, norm, options)

    def _fallback_integrate(self, limits, norm, options):
        bincounts = self._call_rel_counts(
            limits, norm=norm
        )  # TODO: fake data? not to integrate limits?
        edges = limits.binning.edges
        return binned_rect_integration(
            counts=bincounts, edges=edges, limits=limits
        )  # TODO: check integral, correct?

    def _auto_integrate(self, limits, norm, options):
        try:
            integral = self._integrate(limits=limits, norm=norm, options=options)
        except NormNotImplemented:
            unnormalized_integral = self._auto_integrate(
                limits=limits, norm=False, options=options
            )
            normalization = self.normalization(norm, options=options)
            integral = unnormalized_integral / normalization
        except MultipleLimitsNotImplemented:
            integrals = []  # TODO: map?
            for sub_limits in limits:
                integrals.append(
                    self._auto_integrate(limits=sub_limits, norm=norm, options=options)
                )
            integral = z.reduce_sum(integrals, axis=0)  # TODO: remove stack?
        return integral

    @deprecated_norm_range
    def ext_integrate(
        self,
        limits: ztyping.LimitsType,
        norm: ztyping.LimitsType = None,
        *,
        options=None,
        norm_range=None,
    ) -> ztyping.XType:
        """Extended integral of the PDF, i.e. the expected counts or the integral scaled by the yield.

        Args:
            limits: |@doc:pdf.integrate.limits| Limits of the integration. |@docend:pdf.integrate.limits|
            norm: |@doc:pdf.integrate.norm| Normalization of the integration.
               By default, this is the same as the default space of the PDF.
               ``False`` means no normalization and returns the unnormed integral. |@docend:pdf.integrate.norm|
            options: |@doc:pdf.integrate.options| Options for the integration.
               Additional options for the integration. Currently supported options are:
               - type: one of (``bins``)
                 This hints that bins are integrated. A method that is vectorizable, non-dynamic and
                 therefore less suitable for complicated functions is chosen. |@docend:pdf.integrate.options|

        Returns:
            Scalar integration value.
        """
        del norm_range  # taken care of by deprecation decorator
        if not self.is_extended:
            raise NotExtendedPDFError
        if options is None:
            options = {}
        norm = self._check_convert_norm(norm)
        limits = self._check_convert_limits(limits)
        return self._call_ext_integrate(limits, norm, options=options)

    @z.function(wraps="model_binned")
    def _call_ext_integrate(self, limits, norm, *, options):
        with suppress(SpecificFunctionNotImplemented):
            return self._auto_ext_integrate(limits, norm, options=options)
        return self._fallback_ext_integrate(limits, norm, options=options)

    def _fallback_ext_integrate(
        self, limits, norm, *, options
    ):  # TODO: rather use pdf?
        bincounts = self._call_counts(
            limits, norm=norm
        )  # TODO: fake data? not to integrate limits?
        edges = limits.binning.edges
        return binned_rect_integration(counts=bincounts, edges=edges, limits=limits)

    def _auto_ext_integrate(self, limits, norm, *, options):
        try:
            integral = self._ext_integrate(limits=limits, norm=norm, options=options)
        except NormNotImplemented:
            unnormalized_integral = self._auto_ext_integrate(
                limits=limits, norm=False, options=options
            )
            normalization = self.ext_normalization(norm, options=options)
            integral = unnormalized_integral / normalization
        except MultipleLimitsNotImplemented:
            integrals = []  # TODO: map?
            for sub_limits in limits:
                integrals.append(
                    self._auto_integrate(limits=sub_limits, norm=norm, options=options)
                )
            integral = z.reduce_sum(integrals, axis=0)  # TODO: remove stack?
        return integral

    @_BinnedPDF_register_check_support(True)
    def _ext_integrate(self, limits, norm, *, options):
        raise SpecificFunctionNotImplemented

    def create_sampler(
        self,
        n: ztyping.nSamplingTypeIn = None,
        limits: ztyping.LimitsType = None,
        fixed_params: bool | list[ZfitParameter] | tuple[ZfitParameter] = True,
    ) -> BinnedSampler:
        """Create a :py:class:`Sampler` that acts as `Data` but can be resampled, also with changed parameters and n.

            If `limits` is not specified, `space` is used (if the space contains limits).
            If `n` is None and the model is an extended pdf, 'extended' is used by default.


        Args:
            n: The number of samples to be generated. Can be a Tensor that will be
                or a valid string. Currently implemented:

                    - 'extended': samples `poisson(yield)` from each pdf that is extended.

            limits: From which space to sample.
            fixed_params: A list of `Parameters` that will be fixed during several `resample` calls.
                If True, all are fixed, if False, all are floating. If a :py:class:`~zfit.Parameter` is not fixed and
                its
                value gets updated (e.g. by a `Parameter.set_value()` call), this will be reflected in
                `resample`. If fixed, the Parameter will still have the same value as the `Sampler` has
                been created with when it resamples.

        Returns:
            :py:class:`~zfit.core.data.BinnedSampler`

        Raises:
            NotExtendedPDFError: if 'extended' is chosen (implicitly by default or explicitly) as an
                option for `n` but the pdf itself is not extended.
            ValueError: if n is an invalid string option.
            InvalidArgumentError: if n is not specified and pdf is not extended.
        """

        if n is None:
            if self.is_extended:
                n = znp.random.poisson(self.get_yield(), size=1)
            else:
                raise ValueError(
                    f"n cannot be None for sampling of {self} or needs to be extended."
                )
        limits = self._check_convert_limits(limits)

        if fixed_params is True:
            fixed_params = list(self.get_params(floating=False))
        elif fixed_params is False:
            fixed_params = []
        elif not isinstance(fixed_params, (list, tuple)):
            raise TypeError("`Fixed_params` has to be a list, tuple or a boolean.")

        def sample_func(n=n):
            n = znp.array(n)
            sample = self._create_sampler_tensor(limits=limits, n=n)
            return sample

        sample_data = BinnedSampler.from_sample(
            sample_func=sample_func,
            n=n,
            obs=limits,
            fixed_params=fixed_params,
            dtype=self.dtype,
        )

        return sample_data

    @z.function(wraps="sampler")
    def _create_sampler_tensor(self, limits, n):
        sample = self._call_sample(n=n, limits=limits)
        return sample

    def sample(
        self, n: int = None, limits: ztyping.LimitsType = None
    ) -> ZfitBinnedData:
        """Draw a random binned sample from the PDF.

        Args:
            n: |@doc:pdf.sample.n| Number of samples to draw.
               For an extended PDF, the argument is optional and will be the
               poisson-fluctuated expected number of events, i.e. the yield. |@docend:pdf.sample.n|
            limits: |@doc:pdf.sample.limits| Limits of the sampling.
               By default, this is the same as the default space of the PDF. |@docend:pdf.sample.limits|

        Returns:
            ``ZfitBinnedData``: Sampled dataset
        """
        if n is None:
            if self.is_extended:
                n = znp.random.poisson(self.get_yield(), size=1)
            else:
                raise ValueError(
                    f"n cannot be None for sampling of {self} or needs to be extended."
                )
        original_limits = limits
        limits = self._check_convert_limits(limits)
        values = self._call_sample(n, limits)
        if not isinstance(values, ZfitBinnedData):
            values = BinnedData.from_tensor(space=limits, values=values, variances=None)
        if isinstance(original_limits, ZfitSpace):
            values = values.with_obs(original_limits)
        return values

    @z.function(wraps="sample")
    def _call_sample(self, n, limits):
        with suppress(SpecificFunctionNotImplemented):
            self._sample(n, limits)
        return self._fallback_sample(n, limits)

    def _fallback_sample(self, n, limits):
        if limits != self.space:
            raise WorkInProgressError(
                "limits different from the default are not yet available."
                " Please open an issue if you need this:"
                " https://github.com/zfit/zfit/issues/new/choose"
            )
        probs = self.rel_counts(limits)
        values = z.random.counts_multinomial(n, probs=probs, dtype=znp.float64)
        return values

    # ZfitMinimalHist implementation
    def values(self, *, var=None):
        """Histogram values that are either the counts or the normalized counts.

        If the PDF is extended, the counts are returned, otherwise the normalized counts are returned.

        Returns:
            ``ZfitBinnedData``: Histogram values
        """
        if var is not None:
            raise RuntimeError("var argument for `values` is not supported in V1")
        if self.is_extended:
            return self.counts(var)
        else:
            return self.rel_counts(var)

    def update_integration_options(self, *args, **kwargs):
        raise RuntimeError("Integration options not available for BinnedPDF")

    def as_func(self, norm_range: ztyping.LimitsType = False):
        raise WorkInProgressError("as_func not yet available for BinnedPDF")

    @property
    def is_extended(self) -> bool:
        return self._yield is not None

    def set_norm(self, norm):
        raise RuntimeError(
            "set_norm should not be used anymore. Create a new PDF with the desired normalization."
        )

    def create_extended(self, yield_: ztyping.ParamTypeInput) -> ZfitPDF:
        raise WorkInProgressError(
            "create_extended not available for BinnedPDF. Use `extended` in the initialization"
            " instead."
        )

    def get_yield(self) -> ZfitParameter | None:
        if not self.is_extended:
            raise NotExtendedPDFError
        return self._yield

    @classmethod
    def register_analytic_integral(
        cls,
        func: Callable,
        limits: ztyping.LimitsType = None,
        priority: int = 50,
        *,
        supports_norm: bool = False,
        supports_multiple_limits: bool = False,
    ):
        raise RuntimeError("analytic integral not available for BinnedPDF")

    @deprecated_norm_range
    def partial_integrate(
        self,
        x: ztyping.XType,
        limits: ztyping.LimitsType,
        *,
        norm=None,
        options=None,
        norm_range: ztyping.LimitsType = None,
    ) -> ztyping.XType:
        raise WorkInProgressError("partial_integrate not yet available for BinnedPDF")

    @classmethod
    def register_inverse_analytic_integral(cls, func: Callable):
        raise RuntimeError(
            "inverse analytic integral not available for BinnedPDF. It's a histogram, it's already "
            "'analytic'"
        )

    @_BinnedPDF_register_check_support(True)
    def _sample(self, n, limits):
        raise SpecificFunctionNotImplemented

    def _copy(self, deep, name, overwrite_params):
        raise WorkInProgressError

    # factor out with unbinned pdf

    @property
    def norm(self):
        norm = self._norm
        if norm is None:
            norm = self.space
        return norm

    @property
    @deprecated(None, "use `norm` instead.")
    def norm_range(self):
        return self.norm

    # TODO: factor out with unbinned pdf

    def _convert_sort_space(
        self,
        obs: ztyping.ObsTypeInput | ztyping.LimitsTypeInput = None,
        axes: ztyping.AxesTypeInput = None,
        limits: ztyping.LimitsTypeInput = None,
    ) -> ZfitSpace | None:
        """Convert the inputs (using eventually ``obs``, ``axes``) to :py:class:`~zfit.ZfitSpace` and sort them
        according to own ``obs``.

        Args:
            obs:
            axes:
            limits:

        Returns:
        """
        if obs is None:  # for simple limits to convert them
            obs = self.obs
        elif not set(obs).intersection(self.obs):
            raise SpaceIncompatibleError(
                "The given space {obs} is not compatible with the obs of the pdfs{self.obs};"
                " they are disjoint."
            )
        space = convert_to_space(obs=obs, axes=axes, limits=limits)

        if self.space is not None:  # e.g. not the first call
            space = space.with_coords(
                self.space, allow_superset=True, allow_subset=True
            )
        return space

    def counts(
        self, x: ztyping.BinnedDataInputType = None, norm: ztyping.NormInputType = None
    ) -> ZfitBinnedData:
        """Calculate the number of events in each bin.

        This is the integrals of the PDF in each bin.

        Args:
            x: |@doc:pdf.binned.counts.x| Data for the binned PDF.
               The returned counts correspond to the binned axis in ``x``. |@docend:pdf.binned.counts.x|
            norm: |@doc:pdf.binned.counts.norm| Normalization of the counts.
               This normalizes the counts so that the actual sum of all counts is
               equal to the yield. |@docend:pdf.binned.counts.norm|

        Returns:
            ZfitBinnedData: A histogram with the number of events in each bin.
        """
        if not self.is_extended:
            raise NotExtendedPDFError
        x = self._convert_input_binned_x(x, none_is_space=True)
        space = (
            x if isinstance(x, ZfitSpace) else x.space
        )  # TODO: split the convert and sort, make Sorter?
        x = x.with_obs(self.space)
        norm = self._check_convert_norm(norm)
        counts = self._call_counts(x, norm)
        return move_axis_obs(self.space, space, counts)

    @z.function(wraps="model_binned")
    def _call_counts(self, x, norm):
        with suppress(SpecificFunctionNotImplemented):
            return self._auto_counts(x, norm)
        return self._fallback_counts(x, norm)

    def ext_normalization(self, norm, *, options=None):
        if not self.is_extended:
            raise NotExtendedPDFError
        if options is None:
            options = {}
        norm = self._check_convert_norm(norm, none_is_error=True)
        return self._call_ext_normalization(norm, options=options)

    def _call_ext_normalization(self, norm, *, options):
        with suppress(SpecificFunctionNotImplemented):
            return self._ext_normalization(norm, options=options)
        # fallback
        return self.normalization(norm, options=options) / self.get_yield()

    def _ext_normalization(self, norm, *, options):
        raise SpecificFunctionNotImplemented

    def _auto_counts(self, x, norm):
        try:
            counts = self._counts(x, norm=norm)
        except NormNotImplemented:
            unnormed_counts = self._counts(x, norm=False)
            normalization = self.normalization(norm)
            counts = unnormed_counts / normalization
        return counts

    def _fallback_counts(self, x, norm):
        return self._auto_rel_counts(x, norm) * self.get_yield()

    @_BinnedPDF_register_check_support(True)
    def _counts(self, x, norm):
        raise SpecificFunctionNotImplemented

    def rel_counts(
        self, x: ztyping.BinnedDataInputType = None, norm: ztyping.NormInputType = None
    ) -> ZfitBinnedData:
        """Calculate the relative number of events in each bin.

        This is the integrals of the PDF in each bin divided by the integral of the PDF over the whole space.
        It is *not* equal to the density but rather a histogram scaled to 1.

        Args:
            x: |@doc:pdf.binned.counts.x| Data for the binned PDF.
               The returned counts correspond to the binned axis in ``x``. |@docend:pdf.binned.counts.x|
            norm: |@doc:pdf.binned.counts.norm| Normalization of the counts.
               This normalizes the counts so that the actual sum of all counts is
               equal to the yield. |@docend:pdf.binned.counts.norm|

        Returns:
            ZfitBinnedData: A histogram with the relative number of events in each bin.
        """
        x = self._convert_input_binned_x(x, none_is_space=True)
        space = (
            x if isinstance(x, ZfitSpace) else x.space
        )  # TODO: split the convert and sort, make Sorter?
        x = x.with_obs(self.space)
        norm = self._check_convert_norm(norm)
        values = self._call_rel_counts(x, norm)
        return move_axis_obs(self.space, space, values)

    @z.function(wraps="model_binned")
    def _call_rel_counts(self, x, norm):
        with suppress(SpecificFunctionNotImplemented):
            return self._auto_rel_counts(x, norm=norm)
        with suppress(SpecificFunctionNotImplemented):
            return self._auto_counts(x, norm=norm) / self.get_yield()
        return self._fallback_rel_counts(x, norm)

    @_BinnedPDF_register_check_support(True)
    def _rel_counts(self, x, norm):
        raise SpecificFunctionNotImplemented

    def _auto_rel_counts(self, x, norm):
        try:
            rel_counts = self._rel_counts(x, norm=norm)
        except NormNotImplemented:
            unnormed_rel_counts = self._rel_counts(x, norm=False)
            normalization = self.normalization(norm)
            rel_counts = unnormed_rel_counts / normalization
        return rel_counts

    def _fallback_rel_counts(self, x, norm):
        density = self._call_pdf(x, norm)
        rel_counts = density * np.prod(self.space.binning.widths, axis=0)
        return rel_counts

    def set_norm_range(self):
        raise RuntimeError("set_norm_range is removed and should not be used anymore.")

    def to_binned(self, space, *, extended=None, norm=None):
        """Convert the PDF to a binned PDF, returns self.

        For compatibility with unbinned PDFs.
        """
        if isinstance(space, ZfitBinning):
            if space != self.space.binning:
                raise ValueError(
                    "The binning of the PDF and the binning of the space must be equal."
                )
        if space != self.space:
            raise ValueError(
                f"Space must be the same as the PDF's space, as {self} is already a binned PDF."
            )
        if extended is not None:
            raise WorkInProgressError(
                "extended is not implemented yet. Create an extended PDF manually."
            )
        if norm is not None:
            raise WorkInProgressError(
                "norm is not implemented yet. Create a pdf with a different norm range manually."
            )
        return self

    def to_unbinned(self):
        """Convert the PDF to an unbinned PDF."""
        from zfit.models.unbinnedpdf import UnbinnedFromBinnedPDF

        return UnbinnedFromBinnedPDF(self, self.space.with_binning(None))


def binned_rect_integration(
    *,
    limits: ZfitSpace,
    edges: Iterable[znp.array] | znp.array,
    counts: znp.array | None = None,
    density: znp.array | None = None,
    axis: Iterable[int] | int | None = None,
) -> znp.array:
    """Integrate a histogram over *limits*.

    This integrator does take into account that limits do not match the edges.

    Args:
        limits: Limits to integrate over. A possible binning is ignored.
        edges: The edges per axis. They should have the shape ``(1,..., 1, n, 1, ..., 1)``, where n is the *ith* axis.
            ``ZfitBinning`` provides this format on the ``edges` attribute.
        counts: Counts of the histogram. This is what most histograms have and is equal to the density multiplied by
            the binwidth.
            Exactly one of counts or density has to be provided.
        density: The density of a histogram is the bincount divided by the binwidth.
            Exactly one of counts or density has to be provided.
        axis: Which axes to integrate over. Defaults to all.

    Returns:
        Integral with shape corresponding to the non-integrated axes (or a scalar in case of all axes integrated).
    """
    edges = convert_to_container(edges)
    if not isinstance(limits, ZfitSpace):
        raise TypeError(f"limits has to be a ZfitSpace, not {limits}.")
    if counts is not None:
        if density is not None:
            raise ValueError("Either specify 'counts' or 'density' but not both.")
        is_density = False
        values = counts
    elif density is not None:
        is_density = True
        values = density
    else:
        raise ValueError("Need to specify either 'counts' or 'density', not None.")
    ndims = z._get_ndims(values)
    # partial = axis is not None and len(axis) < ndims
    if axis is not None:
        axis = convert_to_container(axis)
        if len(axis) > ndims:
            raise ValueError(
                f"axis {axis} is larger than values has ndims {values.shape}."
            )
    else:
        axis = list(range(ndims))

    scaled_edges, (lower_bins, upper_bins), unscaled_edges = cut_edges_and_bins(
        edges=edges, limits=limits, axis=axis, unscaled=True
    )

    values_cut = tf.slice(
        values, lower_bins, (upper_bins - lower_bins)
    )  # since limits are inclusive

    rank = values.shape.rank
    binwidths = []
    if not is_density:
        binwidths_unscaled = []
    # calculate the binwidth in each dimension
    for i, edge in enumerate(scaled_edges):
        edge_lower_index = [0] * rank
        # int32 is needed! Otherwise the gradient will fail
        edge_lowest_index = znp.array(edge_lower_index, dtype=znp.int32)

        edge_lower_index[i] = 1
        edge_lower_index = znp.array(edge_lower_index, dtype=znp.int32)
        edge_upper_index = [1] * rank
        edge_highest_index = edge_upper_index.copy()
        len_edge = tf.shape(edge)[i]
        edge_highest_index[i] = len_edge
        edge_highest_index = znp.asarray(edge_highest_index, dtype=znp.int32)
        edge_upper_index[i] = len_edge - 1  # len n -> index max is n - 1

        edge_upper_index = znp.asarray(edge_upper_index, dtype=znp.int32)
        lower_edge = tf.slice(
            edge, edge_lowest_index, (edge_upper_index - edge_lowest_index)
        )
        upper_edge = tf.slice(
            edge, edge_lower_index, (edge_highest_index - edge_lower_index)
        )
        binwidths.append(upper_edge - lower_edge)

        if not is_density:
            # unscaled edges to get the ratio
            lower_edge_unscaled = tf.slice(
                unscaled_edges[i],
                edge_lowest_index,
                (edge_upper_index - edge_lowest_index),
            )
            upper_edge_unscaled = tf.slice(
                unscaled_edges[i],
                edge_lower_index,
                (edge_highest_index - edge_lower_index),
            )
            binwidths_unscaled.append(upper_edge_unscaled - lower_edge_unscaled)

    binareas = reduce(
        operator.mul, binwidths
    )  # needs to be python as znp or tf can't broadcast otherwise
    if not is_density:  # scale the counts by the fraction. This is mostly one.
        binareas_uncut = reduce(operator.mul, binwidths_unscaled)
        # binareas_uncut = znp.prod(binwidths_unscaled, axis=0)
        binareas /= binareas_uncut
    values_cut *= binareas
    integral = tf.reduce_sum(values_cut, axis=axis)
    return integral


@z.function(wraps="tensor", keepalive=True)
def cut_edges_and_bins(
    edges: Iterable[znp.array], limits: ZfitSpace, axis=None, unscaled=None
) -> tuple[list[znp.array], tuple[znp.array, znp.array], list | None]:
    """Cut the *edges* according to *limits* and calculate the bins inside.

    The edges within limits are calculated and returned together with the corresponding bin indices. The indices
    mark the lowest and the highest index of the edges that are returned. Additionally, the unscaled edges are returned.

    If the limits are between two edges, this will be treated as the new edge. If the limits are outside the edges,
    all edges in this direction will be returned (but not extended to the limit). For example:

    [0, 0.5, 1., 1.5, 2.] and the limits (0.8, 3.) will return [0.8, 1., 1.5, 2.], ([1], [4])

    .. code-block::

        cut_edges_and_bins([[0., 0.5, 1., 1.5, 2.]], ([[0.8]], [[3]]))



    Args:
        edges: Iterable of tensor-like objects that describe the edges of a histogram. Every object should have rank n
            (where n is the length of *edges*) but only have the dimension i filled out. These are
            tensors that are ready to be broadcasted together.
        limits: The limits that will be used to confine the edges


    Returns:
        edges, (lower bins, upper bins), unscaled_edges:  The edges and the bins are returned.
            The upper bin number corresponds to
            the highest bin which was still (partially) inside the limits **plus one** (so it's the index of the
            edge that is right outside). The unscaled edges are like *edges* but the last edge is the edge
            that is lying not inside anymore, so the actual edge of the last bin number returend.
            This can be used to determine the fraction cut away.
    """
    if axis is not None:
        axis = convert_to_container(axis)
    if unscaled is None:
        unscaled = False
    if unscaled:
        cut_unscaled_edges = []
    else:
        cut_unscaled_edges = None
    cut_scaled_edges = []

    all_lower_bins = []
    all_upper_bins = []
    if isinstance(limits, ZfitSpace):
        lower, upper = limits.limits
    else:
        lower, upper = limits
        lower = znp.asarray(lower)
        upper = znp.asarray(upper)
    lower_all = lower[0]
    upper_all = upper[0]
    rank = len(edges)
    current_axis = 0
    for i, edge in enumerate(edges):
        edge = znp.asarray(edge)
        edge = znp.reshape(edge, (-1,))
        if axis is None or i in axis:
            lower_i = lower_all[current_axis, None]
            edge_minimum = edge[0]
            # edge_minimum = tf.gather(edge, indices=0, axis=i)
            lower_i = znp.maximum(lower_i, edge_minimum)
            upper_i = upper_all[current_axis, None]
            edge_maximum = edge[-1]
            # edge_maximum = tf.gather(edge, indices=tf.shape(edge)[i] - 1, axis=i)
            upper_i = znp.minimum(upper_i, edge_maximum)
            # we get the bins that are just one too far. Then we update this whole bin tensor with the actual edge.
            # The bins index is the index below the value.
            lower_bin_float = tfp.stats.find_bins(
                lower_i, edge, extend_lower_interval=True, extend_upper_interval=True
            )
            lower_bin = tf.reshape(tf.cast(lower_bin_float, dtype=znp.int32), [-1])
            # lower_bins = tf.tensor_scatter_nd_update(zero_bins, [[i]], lower_bin)
            # +1 below because the outer bin is searched, meaning the one that is higher than the value

            upper_bin_float = tfp.stats.find_bins(
                upper_i, edge, extend_lower_interval=True, extend_upper_interval=True
            )
            upper_bin = tf.reshape(tf.cast(upper_bin_float, dtype=znp.int32), [-1]) + 1
            size = upper_bin - lower_bin
            new_edge = tf.slice(
                edge, lower_bin, size + 1
            )  # +1 because stop is exclusive
            new_edge = tf.tensor_scatter_nd_update(
                new_edge, [tf.constant([0]), size], [lower_i[0], upper_i[0]]
            )

            if unscaled:
                new_edge_unscaled = tf.slice(
                    edge, lower_bin, size + 1
                )  # +1 because stop is exclusive

            current_axis += 1
        else:
            lower_bin = [0]
            upper_bin = znp.asarray([edge.shape[0] - 1], dtype=znp.int32)
            new_edge = edge
            if unscaled:
                new_edge_unscaled = edge
        new_shape = [1] * rank
        new_shape[i] = -1
        new_edge = znp.reshape(new_edge, new_shape)
        all_lower_bins.append(lower_bin)
        all_upper_bins.append(upper_bin)
        cut_scaled_edges.append(new_edge)
        if unscaled:
            new_edge_unscaled = znp.reshape(new_edge_unscaled, new_shape)
            cut_unscaled_edges.append(new_edge_unscaled)

    # partial = axis is not None and len(axis) < rank
    #
    # if partial:
    #     scaled_edges_full = list(edges)
    #     for edge, ax in zip(cut_scaled_edges, axis):
    #         scaled_edges_full[ax] = edge
    #     scaled_edges = scaled_edges_full
    #     indices = tf.convert_to_tensor(axis)[:, None]
    #     lower_bins = tf.scatter_nd(indices, lower_bins, shape=(ndims,))
    #     upper_bins = tf.tensor_scatter_nd_update(tf.convert_to_tensor(values.shape),
    #                                              indices, upper_bins)
    # lower_bins_indices = tf.stack([lower_bins, dims], axis=-1)
    # upper_bins_indices = tf.stack([upper_bins, dims], axis=-1)
    # all_lower_bins = tf.cast(znp.sum(all_lower_bins, axis=0), dtype=znp.int32)
    all_lower_bins = tf.concat(all_lower_bins, axis=0)
    all_upper_bins = tf.concat(all_upper_bins, axis=0)
    return cut_scaled_edges, (all_lower_bins, all_upper_bins), cut_unscaled_edges



#  Copyright (c) 2023 zfit

import copy

import numpy as np
import pandas as pd
import pytest
import tensorflow as tf
import uproot

import zfit

obs1 = ("obs1", "obs2", "obs3")

example_data1 = np.random.random(size=(7, len(obs1)))


@pytest.fixture
def obs3d():
    return obs1


@pytest.fixture
def np_data1(obs3d):
    return np.random.random(size=(7, len(obs3d)))


class TestHashPDF(zfit.pdf.BasePDF):
    def __init__(self, obs, lasthash):
        super().__init__(obs=obs)
        self.lasthash = lasthash

    def _unnormalized_pdf(self, x):
        import zfit.z.numpy as znp

        self.lasthash = x.hashint
        return znp.abs(x.unstack_x()[0])


@pytest.fixture
def data1(obs3d, np_data1):
    return zfit.Data.from_numpy(obs=obs3d, array=np_data1)


@pytest.fixture
def space2d():
    obs = ["obs1", "obs2"]
    lower1, lower2 = -1, -2
    upper1, upper2 = 1.9, 3.6
    space1 = zfit.Space(obs[0], limits=(lower1, upper1))
    space2 = zfit.Space(obs[1], limits=(lower2, upper2))
    space2d = space1 * space2
    return space2d


@pytest.mark.parametrize("obs_alias", [None, {"pt1": "pt2", "pt2": "pt1"}])
def test_from_root_limits(obs_alias):
    from skhep_testdata import data_path

    path_root = data_path("uproot-Zmumu.root")

    branches = ["pt1", "pt2"]
    weight_branch = "eta1"
    with uproot.open(path_root) as f:
        tree = f["events"]
        true_data_uncut = tree.arrays(branches + [weight_branch], library="pd")
    lower1 = 40.5
    upper1 = 60.2
    lower2 = 10.5
    upper2 = 40.4
    if obs_alias is not None:  # swap branches in true_data_uncut
        true_data_uncut = true_data_uncut.rename(
            {"pt1": "pt2", "pt2": "pt1"}, axis="columns"
        )

    true_data = true_data_uncut.query(
        f"pt1 > {lower1} & pt1 < {upper1} & pt2 > {lower2} & pt2 < {upper2}"
    )
    true_weights = true_data.pop(weight_branch)
    obs1 = zfit.Space("pt1", limits=(lower1, upper1))
    obs2 = zfit.Space("pt2", limits=(lower2, upper2))
    obs = obs1 * obs2

    data = zfit.Data.from_root(
        path=path_root,
        treepath="events",
        obs=obs,
        weights=weight_branch,
        obs_alias=obs_alias,
    )
    x_np = data.value().numpy()
    np.testing.assert_allclose(x_np, true_data[branches].values)

    weights_np = data.weights.numpy()
    np.testing.assert_allclose(weights_np, true_weights)


@pytest.mark.parametrize(
    "weights_factory",
    [
        lambda: None,
        lambda: 2.0 * tf.ones(shape=(1000,), dtype=tf.float64),
        lambda: np.random.normal(size=1000),
        lambda: "eta1",
    ],
)
def test_from_root(weights_factory):
    weights = weights_factory()

    from skhep_testdata import data_path

    path_root = data_path("uproot-Zmumu.root")

    branches = ["pt1", "pt2", "phi2"]
    with uproot.open(path_root) as f:
        tree = f["events"]
        true_data = tree.arrays(library="pd")

    data = zfit.Data.from_root(
        path=path_root, treepath="events", obs=branches, weights=weights
    )
    x = data.value()
    x_np = x.numpy()
    if weights is not None:
        weights_np = data.weights.numpy()
    else:
        weights_np = weights
    np.testing.assert_allclose(x_np, true_data[branches].values)
    if weights is not None:
        true_weights = (
            weights if not isinstance(weights, str) else true_data[weights].values
        )
        if isinstance(true_weights, tf.Tensor):
            true_weights = true_weights.numpy()
        np.testing.assert_allclose(weights_np, true_weights)
    else:
        assert weights_np is None


@pytest.mark.parametrize(
    "weights_factory",
    [
        lambda: None,
        lambda: 2.0 * tf.ones(shape=(1000,), dtype=tf.float64),
        lambda: np.random.normal(size=1000),
    ],
)
def test_from_numpy(weights_factory, obs3d):
    weights = weights_factory()

    example_data = np.random.random(size=(1000, len(obs1)))
    data = zfit.Data.from_numpy(obs=obs3d, array=example_data, weights=weights)
    x = data.value()
    weights_from_data = data.weights
    if weights_from_data is not None:
        weights_from_data = weights_from_data.numpy()
    x_np = x
    np.testing.assert_array_equal(example_data, x_np)
    if isinstance(weights, tf.Tensor):
        weights = weights.numpy()
    if weights is not None:
        np.testing.assert_allclose(weights_from_data, weights)
    else:
        assert weights_from_data is None


def test_from_to_pandas(obs3d):
    dtype = np.float32
    example_data_np = np.random.random(size=(1000, len(obs3d)))
    example_weights = np.random.random(size=(1000,))
    example_data = pd.DataFrame(data=example_data_np, columns=obs3d)
    data = zfit.Data.from_pandas(obs=obs3d, df=example_data, dtype=dtype)
    x = data.value()
    assert x.dtype == dtype
    x_np = x.numpy()
    assert x_np.dtype == dtype
    np.testing.assert_array_equal(example_data_np.astype(dtype=dtype), x_np)

    # test auto obs retreavel
    example_data2 = pd.DataFrame(data=example_data_np, columns=obs3d)
    data2 = zfit.Data.from_pandas(df=example_data2)
    assert data2.obs == obs3d
    x2 = data2.value()
    x_np2 = x2.numpy()
    np.testing.assert_array_equal(example_data_np, x_np2)
    assert len(data2.obs) == len(obs3d)
    assert len(data2.obs) == len(data2.to_pandas().columns)

    data2w = zfit.Data.from_pandas(df=example_data2, weights=example_weights)
    df2w = data2w.to_pandas()
    data2w2 = zfit.Data.from_pandas(df=df2w)
    assert data2w2.obs == obs3d
    np.testing.assert_allclose(data2w2.weights.numpy(), example_weights)
    np.testing.assert_allclose(data2w2.value().numpy(), example_data_np)
    df = data2.to_pandas()
    assert all(df == example_data)


@pytest.mark.parametrize("weights_as_branch", [True, False])
def test_from_pandas_limits(weights_as_branch, obs3d):
    from skhep_testdata import data_path

    path_root = data_path("uproot-Zmumu.root")

    branches = ["pt1", "pt2"]
    weight_branch = "eta1"
    with uproot.open(path_root) as f:
        tree = f["events"]
        true_data_uncut = tree.arrays(branches + [weight_branch], library="pd")
    lower1 = 23.5
    upper1 = 49.2
    lower2 = 21.5
    upper2 = 44.4
    true_data = true_data_uncut.query(
        f"pt1 > {lower1} & pt1 < {upper1} & pt2 > {lower2} & pt2 < {upper2}"
    )

    true_weights = true_data.pop(weight_branch)
    obs3d = zfit.Space("pt1", limits=(lower1, upper1))
    obs2 = zfit.Space("pt2", limits=(lower2, upper2))
    obs = obs3d * obs2

    weights_for_pandas = (
        true_data_uncut[weight_branch] if weights_as_branch else weight_branch
    )
    data = zfit.Data.from_pandas(
        df=true_data_uncut, obs=obs, weights=weights_for_pandas
    )
    x = data.value()
    x_np = x.numpy()
    np.testing.assert_allclose(x_np, true_data[branches].values)

    weights_np = data.weights.numpy()
    np.testing.assert_allclose(weights_np, true_weights)


@pytest.mark.parametrize(
    "weights_factory",
    [
        lambda: None,
        lambda: 2.0 * tf.ones(shape=(1000,), dtype=tf.float64),
        lambda: np.random.normal(size=1000),
    ],
)
def test_from_tensors(weights_factory):
    weights = weights_factory()
    true_tensor = 42.0 * tf.ones(shape=(1000, 1), dtype=tf.float64)
    data = zfit.Data.from_tensor(obs="obs1", tensor=true_tensor, weights=weights)

    weights_data = data.weights
    x = data.value()
    x_np = x.numpy()
    if isinstance(weights, tf.Tensor):
        weights = weights.numpy()

    np.testing.assert_allclose(x_np, true_tensor.numpy())
    if weights is not None:
        weights_data = weights_data.numpy()
        np.testing.assert_allclose(weights_data, weights)
    else:
        assert weights is None

    df = data.to_pandas()
    data_new = zfit.Data.from_pandas(df=df)
    assert data_new.obs == data.obs
    assert np.allclose(data_new.value().numpy(), data.value().numpy())
    if weights is not None:
        assert np.allclose(data_new.weights.numpy(), data.weights.numpy())
    else:
        assert data_new.weights is None


def test_overloaded_operators(data1):
    with pytest.raises(TypeError):
        a = data1 * 5.0


def test_sort_by_obs(data1, obs3d):
    new_obs = (obs3d[1], obs3d[2], obs3d[0])
    example_data1 = data1.value().numpy()
    new_array = copy.deepcopy(example_data1)[:, np.array((1, 2, 0))]
    # new_array = np.array([new_array[:, 1], new_array[:, 2], new_array[:, 0]])
    assert data1.obs == obs3d, "If this is not True, then the test will be flawed."
    with data1.sort_by_obs(new_obs):
        assert data1.obs == new_obs
        np.testing.assert_array_equal(new_array, data1.value().numpy())
        new_array2 = copy.deepcopy(new_array)[:, np.array((1, 2, 0))]
        # new_array2 = np.array([new_array2[:, 1], new_array2[:, 2], new_array2[:, 0]])
        new_obs2 = (new_obs[1], new_obs[2], new_obs[0])
        with data1.sort_by_obs(new_obs2):
            assert data1.obs == new_obs2
            np.testing.assert_array_equal(new_array2, data1.value().numpy())

        assert data1.obs == new_obs

    assert data1.obs == obs3d
    np.testing.assert_array_equal(example_data1, data1.value().numpy())


def test_data_axis_access(obs3d, data1):
    import zfit.z.numpy as znp

    true_mapping = {
        obs: znp.reshape(arr, (-1, 1))
        for obs, arr in zip(obs3d, data1.value().numpy().T)
    }
    for obs in obs3d:
        np.testing.assert_allclose(data1.value(obs), true_mapping[obs][:, 0])
        np.testing.assert_allclose(data1.value([obs]), true_mapping[obs])
        np.testing.assert_allclose(data1[obs], true_mapping[obs][:, 0])
        np.testing.assert_allclose(data1[[obs]], true_mapping[obs])
    obs2d = [obs3d[2], obs3d[1]]
    array2d = znp.concatenate([true_mapping[obs] for obs in obs2d], axis=1)
    np.testing.assert_allclose(data1[obs2d], array2d)
    np.testing.assert_allclose(data1.value(obs2d), array2d)


def test_subdata(obs3d, data1):
    new_obs = (obs3d[0], obs3d[1])
    new_array = copy.deepcopy(data1.value().numpy())[:, np.array((0, 1))]
    # new_array = np.array([new_array[:, 0], new_array])
    with data1.sort_by_obs(obs=new_obs):
        assert data1.obs == new_obs
        np.testing.assert_array_equal(new_array, data1.numpy())
        new_array2 = copy.deepcopy(new_array)[:, 1]
        # new_array2 = np.array([new_array2[:, 1]])
        new_obs2 = (new_obs[1],)
        with data1.sort_by_obs(new_obs2):
            assert data1.obs == new_obs2
            np.testing.assert_array_equal(new_array2, data1.value().numpy()[:, 0])

            with pytest.raises(ValueError):
                with data1.sort_by_obs(obs=new_obs):
                    data1.value().numpy()

    assert data1.obs == obs3d
    np.testing.assert_array_equal(data1.value().numpy(), data1.value())


@pytest.mark.parametrize(
    "weights_factory",
    [
        lambda: None,
        lambda: np.random.normal(size=5),
    ],
)
def test_data_range(weights_factory):
    data1 = np.array([[1.0, 2], [0, 1], [-2, 1], [-1, -1], [-5, 10]])
    # data1 = data1.transpose()
    obs = ["obs1", "obs2"]
    lower1, lower2 = (0.5, 1), (-3, -2)
    upper1, upper2 = (1.5, 2.5), (-1.5, 1.5)
    space1 = zfit.Space(obs, limits=(lower1, upper1))
    space2 = zfit.Space(obs, limits=(lower2, upper2))
    data_range = space1 + space2
    cut_data1 = data1[np.array((0, 2)), :]

    weights = weights_factory()
    if weights is not None:
        cut_weights = weights[np.array((0, 2))]

    dataset = zfit.Data.from_tensor(obs=obs, tensor=data1, weights=weights)
    value_uncut = dataset.value()
    np.testing.assert_equal(data1, value_uncut.numpy())
    with dataset.set_data_range(data_range):
        value_cut = dataset.value()
        np.testing.assert_equal(cut_data1, value_cut.numpy())
        if dataset.has_weights:
            np.testing.assert_equal(cut_weights, dataset.weights.numpy())
        np.testing.assert_equal(
            data1, value_uncut.numpy()
        )  # check  that the original did NOT change

    np.testing.assert_equal(cut_data1, value_cut.numpy())
    np.testing.assert_equal(data1, dataset.value().numpy())


def test_multidim_data_range():
    nevents1 = 11
    neventstrue1 = 6
    data1 = np.linspace((0, 5), (10, 15), num=nevents1)
    data_true = np.linspace(10, 15, num=neventstrue1)
    lower = ((5, 5),)
    upper = ((10, 15),)
    obs1 = "x"
    obs2 = "y"
    data_range = zfit.Space([obs1, obs2], limits=(lower, upper))
    dataset = zfit.Data.from_numpy(array=data1, obs=data_range)
    xdata = dataset.unstack_x("x")
    assert tf.is_tensor(xdata)
    xdatalist = dataset.unstack_x("x", always_list=True)
    assert len(xdatalist) == 1
    assert isinstance(xdatalist, list)
    assert tf.is_tensor(xdatalist[0])
    yxdata = dataset.unstack_x(["y", "x"])
    assert len(yxdata) == 2
    assert tf.is_tensor(yxdata[0])
    assert xdata.shape == (neventstrue1,)
    assert dataset.nevents.numpy() == 6
    with dataset.sort_by_obs(obs=obs1):
        assert dataset.nevents.numpy() == 6
    assert dataset.nevents.numpy() == 6
    with dataset.sort_by_obs(obs=obs2):
        assert dataset.nevents.numpy() == 6
        np.testing.assert_allclose(data_true, dataset.unstack_x().numpy())

    data2 = np.linspace((0, 5), (10, 15), num=11)[:, 1]
    data_range = zfit.Space([obs1], limits=(5, 15))
    dataset = zfit.Data.from_numpy(array=data2, obs=data_range)
    assert dataset.nevents.numpy() == 11


def test_data_hashing(space2d):
    npdata1 = np.random.uniform(size=(3352, 2))
    # data1 = data1.transpose()
    data1 = zfit.Data.from_numpy(obs=space2d, array=npdata1)
    assert data1.hashint is not None
    testhashpdf = TestHashPDF(obs=space2d, lasthash=data1.hashint)
    assert testhashpdf.lasthash == data1.hashint
    oldhashint = data1.hashint
    data1.set_weights(np.random.uniform(size=data1.nevents))
    assert oldhashint != data1.hashint
    assert data1.hashint != testhashpdf.lasthash
    assert oldhashint == testhashpdf.lasthash
    testhashpdf.pdf(data1, norm=False)
    assert oldhashint != testhashpdf.lasthash
    assert data1.hashint == testhashpdf.lasthash

    with zfit.run.set_graph_mode(
        True
    ):  # meaning integration is now done in graph and has "None"
        oldhashint = data1.hashint
        data1.set_weights(np.random.uniform(size=data1.nevents))
        testhashpdf.pdf(data1)
        assert oldhashint != testhashpdf.lasthash
        assert None == testhashpdf.lasthash


def test_hashing_resample(space2d):
    n = 1534
    pdf = zfit.pdf.Gauss(obs=space2d.with_obs(space2d.obs[0]), mu=0.4, sigma=0.8)
    sample = pdf.create_sampler(n)
    assert sample.hashint is None
    sample.resample()
    assert sample.hashint is None
